#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <complex.h>
#include "cconfig.h"

#define _COL  3
#define _NDIM 4
#define _NSITE   (_NTX*_NTY*_NTZ*_NTT)
#define _MAXBUF  (_COL*_COL*_NDIM*_NTX*_NTY*_NTZ*_NTT)


unsigned long ByteSwap1 (unsigned long nLongNumber)
{
   union u {unsigned long vi; unsigned char c[sizeof(unsigned long)];}; 
   union v {unsigned long ni; unsigned char d[sizeof(unsigned long)];};
   union u un; 
   union v vn; 
   un.vi = nLongNumber; 
   vn.d[0]=un.c[7]; 
   vn.d[1]=un.c[6]; 
   vn.d[2]=un.c[5]; 
   vn.d[3]=un.c[4]; 
   vn.d[4]=un.c[3]; 
   vn.d[5]=un.c[2]; 
   vn.d[6]=un.c[1]; 
   vn.d[7]=un.c[0]; 
   return (vn.ni); 
}


//
// convert endian for ILDG format link data
//
int main(int argc, char *argv[]) 
{
  double complex *ildg_u;
  union dd {double db; unsigned long ul;};
  FILE *fp;

  size_t data_size = sizeof(double complex)*_MAXBUF;

  
  if( argc < 3 ) { 
    fprintf(stderr, "Usage: %s <input file name>  <output file name>\n", argv[0]);
    fprintf(stderr, "This converts the data endian.\n\n");
    return EXIT_FAILURE;
  }

  printf("sizeof(link u)=%ld\n",data_size);

  ildg_u = _mm_malloc(data_size,16);

  /* Read file */
  fp = fopen(argv[1], "r");
  if(fp == (FILE *)NULL) { 
    fprintf(stderr,"Unable to open file %s for reading\n", argv[1]);
    return EXIT_FAILURE;
  }
  fread(ildg_u,data_size,1,fp);
  fclose(fp);

#pragma omp parallel for
  for (int it=0;it<_NTT;it++) {
  for (int iz=0;iz<_NTZ;iz++) {
  for (int iy=0;iy<_NTY;iy++) {
  for (int ix=0;ix<_NTX;ix++) {
    for (int mu=0;mu<_NDIM;mu++) {
      for (int ic=0;ic<_COL;ic++) {
      for (int jc=0;jc<_COL;jc++) {
        
        unsigned long index = jc 
                            + ic*_COL
                            + mu*_COL*_COL 
                            + ix*_COL*_COL*_NDIM 
                            + iy*_COL*_COL*_NDIM*_NTX 
                            + iz*_COL*_COL*_NDIM*_NTX*_NTY
                            + it*_COL*_COL*_NDIM*_NTX*_NTY*_NTZ;
        union dd vre,vim;
        vre.db = creal(ildg_u[index]);
        vim.db = cimag(ildg_u[index]);

        unsigned long ire,iim;
        ire = vre.ul;
        iim = vim.ul;

        vre.ul = ByteSwap1(ire);
        vim.ul = ByteSwap1(iim);

        ildg_u[index] = vre.db + vim.db*I;

      }
      }
    }
  }
  }
  }
  }



  /* Write file */
  fp = fopen(argv[2], "w");
  if(fp == (FILE *)NULL) { 
    fprintf(stderr,"Unable to open file %s for reading\n", argv[2]);
    return EXIT_FAILURE;
  }
  fwrite(ildg_u,data_size,1,fp);
  fclose(fp);

  _mm_free(ildg_u);

  return EXIT_SUCCESS;
}
