#include "configF90.h"

!-----------------------------------------------
!------- Wilson Loop Measurement Class ---------
!
! not completed yet (May, 2011, M.H.)
!

module wilson_loop_class
  use constants_module
!--- module "param_size" was defined in HMCLIB/chrolog_class.F
  use param_size
  use error_class
  use bresenham_module
  use path_generate_class
  use path_product_class
!  use ape_smear_class
  implicit none
  private

  public :: WilsonLoop_set, WilsonLoop_array, WilsonLoopMeas
  public :: new
  public :: delete
  public :: read
  public :: write
  public :: print
  public :: prep
  public :: run
  public :: get_fname ! placed here for debug purpose
  public :: generate ! placed here for debug purpose

! not a good idea
  integer, parameter :: io_unit_wilson_loop_class = 88
!--- index convention, x <-> 1, y <-> 2, z <-> 3, t <-> 4
  integer, parameter :: index_t = 4
!--- header of output file name of results
  character(*), parameter :: fheader = "wloop"
  character(*), parameter :: tlen_header = "tlen"

  integer, parameter :: degit_conf = 4 ! 4 or 5
  integer, parameter :: degit_tlen = 2 ! 2 or 3


  type WilsonLoop_set
    ! R2 = Lmax^2 + Lmid^2 + Lmin^2
    integer :: R2
    ! List of profile_t objects with a given R^2
    type(profile_t), allocatable :: pr(:)
  end type WilsonLoop_set

  type WilsonLoop_array
    ! R(min)^2
    integer :: Rmin2 = 0
    ! R(max)^2
    integer :: Rmax2 = 0
    ! array of WilsonLoop_set objects with a specified R^2
    type(WilsonLoop_set), allocatable :: ws(:)
  end type WilsonLoop_array

  type WilsonLoopMeas
!!$    ! APE smearing (relevant only if flag_ape == 1)
!!$    type(ape_smear_t) :: ape_smear_obj
!!$    ! the following flag must be set equal to 1 if you adopt ape smearing
!!$    ! default is off (0)
!!$    integer :: flag_ape = 0
    ! Rmin^2, Rmax^2 and all displacement vectors
    type(WilsonLoop_array) :: wa
    ! minimum length along t-direction for loops
    integer :: tlen_min
    ! maximum length along t-direction for loops
    integer :: tlen_max
    ! step size of length along t-direction for loops
    integer :: tlen_step
    ! t-coordinate of the starting point of the first loop
    integer :: t0
    ! step size of t-coordinate of the starting point
    integer :: t_step
    ! spatial coordinates of the starting point of the first loop
    integer :: x0(3)
    ! shift of spatial coodinates of the starting point
    integer :: dx(3)
    ! print out the results separately for each displacement vectors
    !   (but averaged over starting points)
    ! default is off (0)
    integer :: flag_print_all_disps = 0
  end type WilsonLoopMeas

!----- member function of WilsonLoop_set ------
  interface new
    module procedure new_WilsonLoop_set
  end interface

  interface delete
    module procedure delete_WilsonLoop_set
  end interface

  interface print
    module procedure print_WilsonLoop_set
  end interface

!----- member function of WilsonLoop_array ------
  interface new
    module procedure new_WilsonLoop_array
  end interface

  interface delete
    module procedure delete_WilsonLoop_array
  end interface

  interface print
    module procedure print_WilsonLoop_array
  end interface

  interface prep
    module procedure prep_WilsonLoop_array
  end interface

!----- member function of WilsonLoopMeas------
  interface new
    module procedure new_WilsonLoopMeas
    module procedure new_WilsonLoopMeas_file
  end interface

  interface delete
    module procedure delete_WilsonLoopMeas
  end interface

  interface read
    module procedure read_WilsonLoopMeas
  end interface

  interface write
    module procedure write_WilsonLoopMeas
    module procedure write_WilsonLoopMeas_file
  end interface

  interface run
    module procedure run_WilsonLoopMeas
  end interface

!--------------- utility ---------------
  interface generate
    module procedure generate_loop_from_spath
  end interface

  interface get_fname
    module procedure get_fname_WilsonLoop
  end interface

contains
!----------- IMPLEMENTATION ------------

subroutine new_WilsonLoop_set( this, R2, prf_size )
!
! inout:
!   this: WilsonLoop_set
!
! input:
!   R2       : R^2
!   prf_size : size of array of profile_t objects
!
  implicit none
  type(WilsonLoop_set), intent(inout) :: this
  integer, intent(in) :: R2
  integer, intent(in) :: prf_size

  this%R2 = R2

  call delete( this )
  allocate( this%pr(prf_size) )

  return
end subroutine new_WilsonLoop_set


subroutine delete_WilsonLoop_set( this )
!
! inout:
!   this: WilsonLoop_set
!
  implicit none
  type(WilsonLoop_set), intent(inout) :: this
!-----------------------------------------------
  integer :: j

  if ( allocated( this%pr ) ) then
    do j = 1, SIZE(this%pr)
      call delete( this%pr(j) )
    enddo
    deallocate( this%pr )
  endif

  return
end subroutine delete_WilsonLoop_set


subroutine print_WilsonLoop_set( this, id_write )
!
! inout:
!   this: WilsonLoop_set
!
! id_write:
!   id_write: unit id associated to the output file
!
  implicit none
  type(WilsonLoop_set), intent(inout) :: this
  integer, intent(in) :: id_write
!-----------------------------------------------
  integer :: j, n

  n = SIZE(this%pr)

  if ( n <= 0 ) return

  if ( nodeid == 0 ) then
    write( id_write, '("### RxR = ", I4, " ###")' ) this%R2
    do j = 1, n
      call print( this%pr(j), id_write )
    enddo
  endif

  return

end subroutine print_WilsonLoop_set

!--------------------------------------------------

subroutine new_WilsonLoop_array( this, Rmin2, Rmax2 )
!
! inout:
!   this: WilsonLoop_array
!
! input:
!   Rmin2 : R(min)^2
!   Rmax2 : R(max)^2
!
  implicit none
  type(WilsonLoop_array), intent(inout) :: this
  integer, intent(in) :: Rmin2, Rmax2
!-----------------------------------------
  integer :: k

!!$  if ( ( this%Rmax2 == Rmax2 ) .and. ( this%Rmin2 == Rmin2 ) ) return
!!$
  call delete( this )

  if ( Rmin2 <= Rmax2 ) then
    this%Rmax2 = Rmax2
    this%Rmin2 = Rmin2
  else
    this%Rmin2 = Rmax2
    this%Rmax2 = Rmin2
  endif

  k = (NTX/2)**2 + (NTY/2)**2 + (NTZ/2)**2
  if ( this%Rmax2 > k ) then
    this%Rmax2 = k
  endif

  allocate( this%ws( this%Rmin2 : this%Rmax2 ) )

  return
end subroutine new_WilsonLoop_array


subroutine delete_WilsonLoop_array( this )
!
! inout:
!   this: WilsonLoop_array
!
  implicit none
  type(WilsonLoop_array), intent(inout) :: this
!--------------------------------------------------
  integer :: R2

  if ( allocated( this%ws ) ) then
    do R2 = this%Rmin2, this%Rmax2
      call delete( this%ws(R2) )
    enddo
    deallocate( this%ws )
  endif

  return
end subroutine delete_WilsonLoop_array


subroutine print_WilsonLoop_array( this, id_write )
!
! inout:
!   this: WilsonLoop_array
!
! id_write:
!   id_write: unit id associated to the output file
!
  implicit none
  type(WilsonLoop_array), intent(inout) :: this
  integer, intent(in) :: id_write
!-----------------------------------------------
  integer :: R2

  if ( nodeid == 0 ) then
    write( id_write, '("# SHAPES of LOOPS ###")' )
    do R2 = this%Rmin2, this%Rmax2
      call print( this%ws(R2), id_write )
    enddo
  endif

  return
end subroutine print_WilsonLoop_array


subroutine prep_WilsonLoop_array( this, da, len_t )
! generate all profile_t objects for loops
!
! inout :
!   this : WilsonLoop_array object
!   da   : disp_set_array object
!          ( da is reinitialized 
!            if da%Rmin2 /= this%Rmin2 or da%Rmax2 /= this%Rmax2 )
!
! input :
!   len_t : length of loop along t-direction 
!
  implicit none
  type(WilsonLoop_array), intent(inout) :: this
  type(disp_set_array), intent(inout), target :: da
  integer, intent(in) :: len_t
!-----------------------------------------------------------
  type(segment_t), target :: loop
  type(disp), pointer :: dp_p => NULL()
  integer :: num_seg
  integer :: num_disp
  integer :: Rmin2, Rmax2, R2, j

  if ( .not. allocated( this%ws ) ) then
    call error_stop("WilsonLoop_set objects have not allocated yet!")
  endif

  Rmin2 = this%Rmin2
  Rmax2 = this%Rmax2

  if ( ( da%Rmin2 /= Rmin2 ) .or. ( da%Rmax2 /= Rmax2 ) ) then
    call new( da, Rmin2, Rmax2 )
    call run( da )
  endif

  do R2 = Rmin2, Rmax2
    num_disp = get_num( da, R2 )
    call new( this%ws(R2), R2, num_disp )
    dp_p => get( da, R2 )
    do j = 1, num_disp
      if ( .not. associated( dp_p ) ) then
        call error_stop("disp object is empty !")
      endif
      call new( loop )
      call generate( loop, num_seg, dp_p, len_t )
      ! make a profile from list of segment_t objects
      call new( this%ws(R2)%pr(j), loop, num_seg )
      dp_p => dp_p%next
    enddo ! j (loop over disps of a give R2)
  enddo ! R2

  return
end subroutine prep_WilsonLoop_array

!-------------------------------------------------------

subroutine new_WilsonLoopMeas( this, prm_unit )
!
! inout :
!   this : WilsonLoopMeas object
!
! input :
!   prm_unit : unit id of parameter file (opotional) 
!
  implicit none
  type(WilsonLoopMeas), intent(inout) :: this
  integer, intent(in), optional :: prm_unit

  call delete( this )

  if ( present(prm_unit) ) then
    call read( this, prm_unit )
  endif

  return
end subroutine new_WilsonLoopMeas


subroutine new_WilsonLoopMeas_file( this, ifname )
!
! inout :
!   this : WilsonLoopMeas object
!
! input :
!   ifname : input file name 
!
  implicit none
  type(WilsonLoopMeas), intent(inout) :: this
  character(LEN=*), intent(in) :: ifname
!--------------------------------------------
  integer, parameter :: id_unit = io_unit_wilson_loop_class

  open( unit=id_unit, file=ifname, status='old',  &
 &      action='read', position='rewind' )
  call new( this, id_unit )  
  close(id_unit)

  return
end subroutine new_WilsonLoopMeas_file


subroutine delete_WilsonLoopMeas( this )
!
! inout :
!   this : WilsonLoopMeas object
!
  implicit none
  type(WilsonLoopMeas), intent(inout) :: this

  call delete( this%wa )

  return
end subroutine delete_WilsonLoopMeas


subroutine read_WilsonLoopMeas( this, prm_unit )
!
! inout :
!   this : WilsonLoopMeas object
!
! input :
!   prm_unit : unit id of parameter file
!
  implicit none
  type(WilsonLoopMeas), intent(inout) :: this
  integer, intent(in) :: prm_unit
!------------------------------------
  integer :: Rmin2, Rmax2

  read(prm_unit,*)

  ! (sptial length)^2
  read(prm_unit,*)
  read(prm_unit,*) Rmin2
  read(prm_unit,*)
  read(prm_unit,*) Rmax2
  call new( this%wa, Rmin2, Rmax2 )

  ! temporal lengths
  read(prm_unit,*)
  read(prm_unit,*) this%tlen_min
  read(prm_unit,*)
  read(prm_unit,*) this%tlen_max
  read(prm_unit,*)
  read(prm_unit,*) this%tlen_step

  ! information on t-coodinates of starting point
  read(prm_unit,*)
  read(prm_unit,*) this%t0
  read(prm_unit,*)
  read(prm_unit,*) this%t_step

  ! information on sptial coodinates of starting point
  read(prm_unit,*)
  read(prm_unit,*) this%x0(1), this%x0(2), this%x0(3)
  read(prm_unit,*)
  read(prm_unit,*) this%dx(1), this%dx(2), this%dx(3)

  read(prm_unit,*)
  read(prm_unit,*) this%flag_print_all_disps

  return
end subroutine read_WilsonLoopMeas


subroutine write_WilsonLoopMeas( this, id_write )
! write working parameters 
!
! input :
!   this : WilsonLoopMeas object
!   id_write : unit id of file to write paraemeters
!
  implicit none
  type(WilsonLoopMeas), intent(in) :: this
  integer, intent(in) :: id_write

  if ( nodeid == 0 ) then
    write(id_write,'("#          R(min)^2 :", I5)') this%wa%Rmin2
    write(id_write,'("#          R(max)^2 :", I5)') this%wa%Rmax2
    write(id_write,'("#         tlen(min) :", I3)') this%tlen_min 
    write(id_write,'("#         tlen(max) :", I3)') this%tlen_max 
    write(id_write,'("#        tlen(step) :", I3)') this%tlen_step 
    write(id_write,'("#                t0 :", I3)') this%t0
    write(id_write,'("#          t0(step) :", I3)') this%t_step 
    write(id_write,'("#        x0  y0  z0 :", 3I4)') &
 &    this%x0(1), this%x0(2), this%x0(3)
    write(id_write,'("#        dx  dy  dz :", 3I4)') &
 &    this%dx(1), this%dx(2), this%dx(3)
    if ( this%flag_print_all_disps /= 0 ) then
      write(id_write,'("# print all disps ? : YES")')
    else
      write(id_write,'("# print all disps ? : NO")')
    endif
  endif

  return
end subroutine write_WilsonLoopMeas


subroutine write_WilsonLoopMeas_file( this, ofname )
! write working parameters 
!
! input :
!   this : WilsonLoopMeas object
!   ofname : output file name
!
  implicit none
  type(WilsonLoopMeas), intent(in) :: this
  character(LEN=*), intent(in) :: ofname
!-------------------------------------------
  integer, parameter :: id_unit = io_unit_wilson_loop_class

  if ( nodeid == 0 ) then
    open( unit=id_unit, file=TRIM(ofname), status='unknown',  &
 &        action='write', form='formatted' )
    call write( this, id_unit )
    close( id_unit )
  endif

  return
end subroutine write_WilsonLoopMeas_file


subroutine generate_loop_from_spath( loop, num_seg, dsp, len_t )
!
! form a loop, (s-path by dsp) -> (+t-path with length len_t)
!              -> (s-path by reserse of dsp) -> (-t-path with length len_t)
! as the list of segment_t objects
!
! output:
!   loop: list of segmenet_t objects
!   num_seg : total number of segments forming the loop
!
! input:
!   dsp   : disp object
!   len_t : length along t-direction
!
  implicit none
  type(segment_t), intent(out), target :: loop
  integer, intent(out) :: num_seg
  type(disp), intent(in) :: dsp
  integer, intent(in) :: len_t
!----------------------------------------
  type(segment_t), pointer :: s_p
  integer :: j
  integer :: len_s

  len_s = SIZE(dsp%dir)
  s_p => loop

!-- form (s-path corresponding to dsp)
  s_p%mu = dsp%dir(1)
  s_p%len = 1
  num_seg = 1
  do j = 2, len_s
    if ( dsp%dir(j) == dsp%dir(j-1) ) then
      s_p%len = s_p%len + 1
    else
      allocate( s_p%next )
      s_p => s_p%next
      s_p%mu = dsp%dir(j)
      s_p%len = 1
      num_seg = num_seg + 1
    endif
  enddo ! j

!-- form (+t)-path
  allocate(s_p%next)
  s_p => s_p%next
  s_p%mu = 4
  s_p%len = len_t
  num_seg = num_seg + 1

!-- form (s-path corresponding to inverse of dsp)
  allocate(s_p%next)
  s_p => s_p%next
  s_p%mu = - dsp%dir(len_s)
  s_p%len = 1
  num_seg = num_seg + 1
  do j = ( len_s - 1 ), 1, -1
    if ( dsp%dir(j) == dsp%dir(j+1) ) then
      s_p%len = s_p%len + 1
    else
      allocate( s_p%next )
      s_p => s_p%next
      s_p%mu = - dsp%dir(j)
      s_p%len = 1
      num_seg = num_seg + 1
    endif
  enddo ! j

!-- form (-t)-path
  allocate(s_p%next)
  s_p => s_p%next
  s_p%mu = -4
  s_p%len = len_t
  num_seg = num_seg + 1

  return
end subroutine generate_loop_from_spath


subroutine get_fname_WilsonLoop( fname, path, iconf, tlen )
! construct name of file to print out result
!
! input:
!   fname : file name to print out result
!
! input:
!   path  : path name
!   iconf : conf id
!   tlen  : length of Wilson loop along t-direction
!
  character(CHARLEN), intent(out) :: fname
  character(len=*), intent(in) :: path
  integer, intent(in) :: iconf
  integer, intent(in) :: tlen
!----------------------------------------------
  character(32) :: ctlen
  character(degit_conf) :: cconf
  integer :: j

  if ( degit_tlen == 2 ) then
    write(ctlen, '(I2)') tlen
  else
    write(ctlen, '(I3)') tlen
  endif
  do j = 1, degit_tlen
    if ( ctlen(j:j) == ' ' ) ctlen(j:j) = '0'
  enddo
  ctlen = TRIM(TRIM(ADJUSTL(tlen_header)) // TRIM(ADJUSTL(ctlen)))

  if ( degit_conf == 4 ) then
    write(cconf, '(I4)') iconf
  else
    write(cconf, '(I5)') iconf
  endif
  do j = 1, degit_conf
    if ( cconf(j:j) == ' ' ) cconf(j:j) = '0'
  enddo

  fname = TRIM( &
 &          TRIM(ADJUSTL(fheader)) // "." // TRIM(ADJUSTL(ctlen))  &
 &          // "." // TRIM(ADJUSTL(cconf)) )
  fname = TRIM(TRIM(ADJUSTL(path)) // "/" // TRIM(ADJUSTL(fname)))

  return
end subroutine get_fname_WilsonLoop


subroutine run_WilsonLoopMeas( this, path, iconf, ue, uo )
!
! perform measurement of Wilson lines
!
! inout :
!   this : WilsonLoopMeas object
!
! input:
!   path  : path to the output file (directory is assumed to exist)
!   iconf : id assigned to conf (used to generate output file name)
!   ue    : link fields on even sites
!   uo    : link fields on odd sites
!
  implicit none
  type(WilsonLoopMeas), intent(inout) :: this
  complex(DP), intent(in) :: ue(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
  complex(DP), intent(in) :: uo(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
  character(LEN=*), intent(in) :: path
  integer, intent(in) :: iconf
!-----------------------------------------------------------------------
  complex(DP) :: wl(COL,COL)
  real(DP) :: avr_all, avr_pos, rtmp
  character(CHARLEN) :: fname
  type(disp_set_array) :: da
  type(path_product_t) :: pp
  integer :: tlen, tlen_max
  integer :: cnt_pos, cnt_all
  integer :: numX, numY, numZ, iterX, iterY, iterZ
  integer :: numT, iterT
  integer :: j
  integer :: R2
  integer :: num_prf
  integer, parameter :: id_write = io_unit_wilson_loop_class

  numX = ( NTX - this%x0(1) ) / this%dx(1) + 1
  if ( numX <= 0 ) call error_stop("numX is not positive !")
  numY = ( NTY - this%x0(2) ) / this%dx(2) + 1
  if ( numY <= 0 ) call error_stop("numY is not positive !")
  numZ = ( NTZ - this%x0(3) ) / this%dx(3) + 1
  if ( numZ <= 0 ) call error_stop("numZ is not positive !")
  numT = ( NTT - this%t0 ) / this%t_step + 1
  if ( numT <= 0 ) call error_stop("numT is not positive !")

  tlen_max = MIN( this%tlen_max, NTT )
  if ( this%tlen_min > tlen_max )  &
 &  call error_stop("minumum of t-length is larger than maximum")

  ! create all spatial displacements
  call new( da, this%wa%Rmin2, this%wa%Rmax2 )
  call run( da )

  tlen = this%tlen_min
  do while ( tlen <= tlen_max )

    call new( pp )
    ! create data of all loops from data of spatial displacements
    call prep( this%wa, da, tlen )

    call get_fname( fname, path, iconf, tlen )
    open( unit=id_write, file=fname, status='unknown', action='write' )
!!$    call print( this%wa, id_write )

    do R2 = this%wa%Rmin2, this%wa%Rmax2

      cnt_all = 0
      avr_all = 0.0_DP

      num_prf = SIZE(this%wa%ws(R2)%pr)

      write( id_write, '("### RxR = ", I4, " ###" )' ) R2
      write( id_write,  &
 &           '("##### number of profiles = ", I3 )' ) num_prf

      do j = 1, num_prf ! loop over profiles with a fixed R2

        ! copy profile information to path_t object member of pp
        call new( pp%path_obj%prf_obj, this%wa%ws(R2)%pr(j) )


        call print( pp%path_obj%prf_obj, id_write )


        ! start measurement, average over starting position
        cnt_pos = 0
        avr_pos = 0.0_DP
        do iterX = 1, numX
          pp%path_obj%x(1) = this%x0(1) + ( iterX - 1 ) * this%dx(1)
          do iterY = 1, numY
            pp%path_obj%x(2) = this%x0(2) + ( iterY - 1 ) * this%dx(2)
            do iterZ = 1, numZ
              pp%path_obj%x(3) = this%x0(3) + ( iterZ - 1 ) * this%dx(3)
              do iterT = 1, numT
                pp%path_obj%x(4) = this%t0 + ( iterT - 1 ) * this%t_step

                call run( pp, wl, ue, uo )
                rtmp = ( REAL( wl(1,1), KIND=DP ) + REAL( wl(2,2), KIND=DP ) ) &
 &                     / REAL( COL, KIND=DP )
                avr_pos = avr_pos + rtmp
                cnt_pos = cnt_pos + 1

              enddo ! iterT
            enddo ! iterZ
          enddo ! iterY
        enddo ! iterX
        avr_pos = avr_pos / REAL( cnt_pos, KIND=DP )

        avr_all = avr_all + avr_pos
        cnt_all = cnt_all + 1

        ! print average over positions for a fixed spatial profile
        if ( this%flag_print_all_disps /= 0 ) then
          if ( num_prf > 1 )  &
 &          write( id_write,  &
 &               '("##### @@@", I3, "th profile: ", E24.16)') j, avr_pos
        endif

      enddo ! end of loop over profiles with a fixed R2

      if ( cnt_all >= 1 ) then
        avr_all = avr_all / REAL( cnt_all, KIND=DP )
        write( id_write, '("### Average : ", E24.16, " @@@", I8 )' )  &
 &             avr_all, cnt_all
      endif

    end do ! R2

    close( id_write )

!!$    call delete( this%wa )
    tlen = tlen + this%tlen_step ! t-length for next measurement

    call delete( pp )
  enddo ! tlen

  call delete( da )

  return
end subroutine run_WilsonLoopMeas


end module wilson_loop_class
