#ifndef _CU_PARAM_SIZE_H
#define _CU_PARAM_SIZE_H

/////////////////////////////////////////
// Constants
/////////////////////////////////////////
#define NDIM 4
#define NX   (_NTX/_NDIMX)
#define NY   (_NTY/_NDIMY)
#define NZ   (_NTZ/_NDIMZ)
#define NT   _NTT
#define NX1 (NX+1)
#define NY1 (NY+1)
#define NZ1 (NZ+1)
#define NT1 (NT+1)
#define NTH (NT/2)

#define COL    3
#define CLCL   (COL*COL)
#define SPIN   4
#define CLSP  (COL*SPIN)
#define CSH   (COL*SPIN/2)
#define CLSPH ((CLSP/2+1)*(CLSP/4))

#define HNSITE (NTH*NZ*NY*NX)

#ifdef __CUDACC__
#define _ALIG_ __align__(16)
#else
#define _ALIG_
#endif

///////////////////////////////////////////////
// Indexing functions
///////////////////////////////////////////////
typedef struct _ALIG_ {
  int t,z,y,x;
} d4site;

typedef struct _ALIG_ {
  d4site thread;
  d4site block;
} bksite;

static __inline__ bksite d4site2bksite(d4site isite) {
//
// convert normal-ordered 4D-site-index to gpu-block-thread 4D-site-index
//
  bksite jsite;
  jsite.thread.t = isite.t % _NBT;
   jsite.block.t = isite.t / _NBT;
  jsite.thread.z = isite.z % _NBZ;
   jsite.block.z = isite.z / _NBZ;
  jsite.thread.y = isite.y % _NBY;
   jsite.block.y = isite.y / _NBY;
  jsite.thread.x = isite.x % _NBX;
   jsite.block.x = isite.x / _NBX;
  return(jsite);
}

static __inline__ d4site bksite2d4site(bksite isite) {
//
// convert gpu-block-thread 4D-site-index to normal-ordered 4D-site-index
//
  d4site jsite;
  jsite.t = isite.thread.t + isite.block.t * _NBT;
  jsite.z = isite.thread.z + isite.block.z * _NBZ;
  jsite.y = isite.thread.y + isite.block.y * _NBY;
  jsite.x = isite.thread.x + isite.block.x * _NBX;
  return(jsite);
}

static __inline__ int bksite2index(bksite isite) {
//
// convert gpu-block-thread 4D-site-index to linear-1D index
//
  return (  isite.thread.t
          + (_NBT)*(isite.thread.z)
          + (_NBT)*(_NBZ)*(isite.thread.y)
          + (_NBT)*(_NBZ)*(_NBY)*(isite.thread.x)
          + (_NBT)*(_NBZ)*(_NBY)*(_NBX)*(isite.block.t)
          + (_NBT)*(_NBZ)*(_NBY)*(_NBX)*(_NGT)*(isite.block.z)
          + (_NBT)*(_NBZ)*(_NBY)*(_NBX)*(_NGT)*(_NGZ)*(isite.block.y)
          + (_NBT)*(_NBZ)*(_NBY)*(_NBX)*(_NGT)*(_NGZ)*(_NGY)*(isite.block.x));
}

static __inline__ int bksite2thread(bksite isite) {
//
// get gpu-thread index from gpu-block-thread 4D-site-index
//
  return (  isite.thread.t
          + (_NBT)*(isite.thread.z)
          + (_NBT)*(_NBZ)*(isite.thread.y)
          + (_NBT)*(_NBZ)*(_NBY)*(isite.thread.x));
}

static __inline__ int bksite2block(bksite isite) {
//
// get gpu-block index from gpu-block-thread 4D-site-index
//
  return ( (isite.block.t)
          +(_NGT)*(isite.block.z)
          +(_NGT)*(_NGZ)*(isite.block.y)
          +(_NGT)*(_NGZ)*(_NGY)*(isite.block.x));
}

static __inline__ bksite index2bksite(int isite) {
//
// convert 1D-index to gpu-block-thread 4D-site-index
//
  bksite jsite;
  jsite.thread.t = isite % _NBT;
  jsite.thread.z = isite /(_NBT) % _NBZ;
  jsite.thread.y = isite /(_NBT*_NBZ) % _NBY;
  jsite.thread.x = isite /(_NBT*_NBZ*_NBY) % _NBX;
  jsite.block.t  = isite /(_NBT*_NBZ*_NBY*_NBX) % _NGT;
  jsite.block.z  = isite /(_NBT*_NBZ*_NBY*_NBX*_NGT) % _NGZ;
  jsite.block.y  = isite /(_NBT*_NBZ*_NBY*_NBX*_NGT*_NGZ) % _NGY;
  jsite.block.x  = isite /(_NBT*_NBZ*_NBY*_NBX*_NGT*_NGZ*_NGY) % _NGX;
  return(jsite);
}

///////////////////////////////////////////////
// my complex definition
///////////////////////////////////////////////
typedef float2 cuComplex;

///////////////////////////////////////////////
// Texturefetch indexing macro
///////////////////////////////////////////////
#define __loadu_tex(turi,ithd,ic,mu,iblk)  tex1Dfetch(turi,(ithd)+THREADS*(ic)+THREADS*COL*(mu)+THREADS*COL*NDIM*(iblk))
#define __loady_tex(tyri,ithd,ics,iblk)    tex1Dfetch(tyri,(ithd)+THREADS*(ics)+THREADS*CSH*(iblk))
#define __loadf_tex(tfri,ithd,ics,iblk)    tex1Dfetch(tfri,(ithd)+THREADS*(ics)+THREADS*CLSPH*(iblk))

///////////////////////////////////////////////
// Field definitions used in GPU
///////////////////////////////////////////////
typedef struct _ALIG_ {
  float4 c[THREADS];
} cufloat4_thread;

///////////////////////
// quark field
///////////////////////
typedef struct _ALIG_ {
  cufloat4_thread yri[CSH];
} cuhqf_thread;

typedef struct _ALIG_ {
  cuhqf_thread b[BLOCKS];
} cuhqfield;

//////////////////////
// gluon(link) field
//////////////////////
typedef struct _ALIG_ {
  cufloat4_thread uri[COL];
} cuhgsf_thread;

typedef struct _ALIG_ {
  cuhgsf_thread v[NDIM];
} cuhgvf_thread;

typedef struct _ALIG_ {
  cuhgvf_thread b[BLOCKS];
} cuhgvfield;

//////////////////////
// inverse 
// clover term field
//////////////////////
typedef struct _ALIG_ {
  cufloat4_thread fri[CLSPH];
} cuclv_thread;

typedef struct _ALIG_ {
  cuclv_thread b[BLOCKS];
} cuhclvfield;

#endif
