static __global__ void 
assign_mult_gamma5_kernel(      cu_wqf_eo_blk *gye,
                          const int4          *blk_indx,
                          const cu_wqf_eo_blk  *ye)
{
////////////////////////////////////////////////////////////////////////////////
// Multiply gamma5 : gye = gamma5 ye
//
//     gye : even/odd site fermion (ieo=0/1) (output)
//      ye : odd/even site fermion (ieo=0/1) (input)
//
////////////////////////////////////////////////////////////////////////////////
  __align__(16) float4 yri[CSH],gyri[CSH];
  unsigned int ithd = threadIdx.x;
  unsigned int iblk = blk_indx[blockIdx.x].w;

  (*gye).igx      = (*ye).igx;
  (*gye).ieo      = (*ye).ieo;
  (*gye).ipeo     = (*ye).ipeo;
  (*gye).ipsite.x = (*ye).ipsite.x;
  (*gye).ipsite.y = (*ye).ipsite.y;
  (*gye).ipsite.z = (*ye).ipsite.z;
  (*gye).ipsite.w = (*ye).ipsite.w;

  yri[0] = (*ye).blk[iblk].thd[0].f[ithd];
  yri[1] = (*ye).blk[iblk].thd[1].f[ithd];
  yri[2] = (*ye).blk[iblk].thd[2].f[ithd];
  yri[3] = (*ye).blk[iblk].thd[3].f[ithd];
  yri[4] = (*ye).blk[iblk].thd[4].f[ithd];
  yri[5] = (*ye).blk[iblk].thd[5].f[ithd];

  __yr11(gyri) = __yr13(yri);
  __yr21(gyri) = __yr23(yri);
  __yr31(gyri) = __yr33(yri);
  __yr12(gyri) = __yr14(yri);
  __yr22(gyri) = __yr24(yri);
  __yr32(gyri) = __yr34(yri);
  __yr13(gyri) = __yr11(yri);
  __yr23(gyri) = __yr21(yri);
  __yr33(gyri) = __yr31(yri);
  __yr14(gyri) = __yr12(yri);
  __yr24(gyri) = __yr22(yri);
  __yr34(gyri) = __yr32(yri);

  __yi11(gyri) = __yi13(yri);
  __yi21(gyri) = __yi23(yri);
  __yi31(gyri) = __yi33(yri);
  __yi12(gyri) = __yi14(yri);
  __yi22(gyri) = __yi24(yri);
  __yi32(gyri) = __yi34(yri);
  __yi13(gyri) = __yi11(yri);
  __yi23(gyri) = __yi21(yri);
  __yi33(gyri) = __yi31(yri);
  __yi14(gyri) = __yi12(yri);
  __yi24(gyri) = __yi22(yri);
  __yi34(gyri) = __yi32(yri);

  (*gye).blk[iblk].thd[0].f[ithd] = gyri[0];
  (*gye).blk[iblk].thd[1].f[ithd] = gyri[1];
  (*gye).blk[iblk].thd[2].f[ithd] = gyri[2];
  (*gye).blk[iblk].thd[3].f[ithd] = gyri[3];
  (*gye).blk[iblk].thd[4].f[ithd] = gyri[4];
  (*gye).blk[iblk].thd[5].f[ithd] = gyri[5];
}

static __global__ void 
accum_mult_gamma5_kernel(const int4    *blk_indx,
                         cu_wqf_eo_blk *ye)
{
////////////////////////////////////////////////////////////////////////////////
// Multiply gamma5 : ye := gamma5 ye
//
//      ye : odd/even site fermion (ieo=0/1) (input/output)
//
////////////////////////////////////////////////////////////////////////////////
  __align__(16) float4 yri[CSH],gyri[CSH];
  unsigned int ithd = threadIdx.x;
  unsigned int iblk = blk_indx[blockIdx.x].w;

  yri[0] = (*ye).blk[iblk].thd[0].f[ithd];
  yri[1] = (*ye).blk[iblk].thd[1].f[ithd];
  yri[2] = (*ye).blk[iblk].thd[2].f[ithd];
  yri[3] = (*ye).blk[iblk].thd[3].f[ithd];
  yri[4] = (*ye).blk[iblk].thd[4].f[ithd];
  yri[5] = (*ye).blk[iblk].thd[5].f[ithd];

  __yr11(gyri) = __yr13(yri);
  __yr21(gyri) = __yr23(yri);
  __yr31(gyri) = __yr33(yri);
  __yr12(gyri) = __yr14(yri);
  __yr22(gyri) = __yr24(yri);
  __yr32(gyri) = __yr34(yri);
  __yr13(gyri) = __yr11(yri);
  __yr23(gyri) = __yr21(yri);
  __yr33(gyri) = __yr31(yri);
  __yr14(gyri) = __yr12(yri);
  __yr24(gyri) = __yr22(yri);
  __yr34(gyri) = __yr32(yri);

  __yi11(gyri) = __yi13(yri);
  __yi21(gyri) = __yi23(yri);
  __yi31(gyri) = __yi33(yri);
  __yi12(gyri) = __yi14(yri);
  __yi22(gyri) = __yi24(yri);
  __yi32(gyri) = __yi34(yri);
  __yi13(gyri) = __yi11(yri);
  __yi23(gyri) = __yi21(yri);
  __yi33(gyri) = __yi31(yri);
  __yi14(gyri) = __yi12(yri);
  __yi24(gyri) = __yi22(yri);
  __yi34(gyri) = __yi32(yri);

  (*ye).blk[iblk].thd[0].f[ithd] = gyri[0];
  (*ye).blk[iblk].thd[1].f[ithd] = gyri[1];
  (*ye).blk[iblk].thd[2].f[ithd] = gyri[2];
  (*ye).blk[iblk].thd[3].f[ithd] = gyri[3];
  (*ye).blk[iblk].thd[4].f[ithd] = gyri[4];
  (*ye).blk[iblk].thd[5].f[ithd] = gyri[5];
}

static void 
assign_mult_gamma5( cu_wqf_eo *gye, const cu_wqf_eo *ye)
{
  (*gye).ieo  = (*ye).ieo;
  (*gye).ipeo = (*ye).ipeo;
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    switch_gpu_(&igx);
    assign_mult_gamma5_kernel<<<_NBLOCKS,_NTHREADS>>>((*gye).g[igx],phys_blk_d.g[igx],(*ye).g[igx]);
    CUT_CHECK_ERROR("assign_mult_gamma5 kernel execution faild");
  }
}

static void 
accum_mult_gamma5( cu_wqf_eo *ye)
{
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    switch_gpu_(&igx);
    accum_mult_gamma5_kernel<<<_NBLOCKS,_NTHREADS>>>(phys_blk_d.g[igx],(*ye).g[igx]);
    CUT_CHECK_ERROR("accum_mult_gamma5 kernel execution faild");
  }
}
