#include <config.h>

!#define _CHECK_TRACE
!#define _CHECK_RAND
!#define _CHECK_GAUGEINV

!
!= SF measurement class
!
!== Version
!
! $Id: sf_measure_class.F90,v 1.5 2011/10/31 07:48:16 ishikawa Exp $
!
module sf_measure_class
!
! This module contains the following SF measurements.
!
!  - SF scheme coupling measurement.
!  - PCAC correlation function measurement with SF boundary. (f_A,f_P,f_A',f_P')
!
  use lattice_class
  use logfile_class
  use timer_class
  use quark_prop_class
  use wavefunc_param_class
  use wavefunc_class
  use field_gauge_class
  use field_fermion_class
#ifdef _CLOVER
  use quark_clover_class
#else
  use quark_wilson_class
#endif
  use hmc_status_class
#ifdef _STOUT
  use hmc_stout_class
#endif
  implicit none
  private
  public :: sf_measurement
  public :: new, delete
  public :: measure

  integer, parameter :: NSPACE=NX*NY*NZ

  type(rand_gfsr_obj), save :: m_rand
  logical, save :: m_is_initialized = .FALSE.

  real(DP), save :: m_theta(NDIM-1)  ! phase angle for generic spatial boundary condition

  integer, parameter :: SF_MEASURE_NONE      = 0
  integer, parameter :: SF_MEASURE_GSF_GLUON = 1
  integer, parameter :: SF_MEASURE_GSF_QUARK = 2
  integer, parameter :: SF_MEASURE_PCAC      = 4
  integer, parameter :: SF_MEASURE_ZFACTOR   = 8

  type random_parameter
  !
  ! randam number paramters
  !
    character(len=CHARLEN) :: parameter_fname
    character(len=CHARLEN) :: seed_fpath
    character(len=CHARLEN) :: seed_fname
    integer :: run_number=0            ! run number
    integer :: random_number_seed=1    ! stochastic trace estimator, noise seed
    integer :: nnoise=1                ! stochastic trace estimator, noise number
    integer :: idummy1=0
  end type

  type sf_measurement
  !
  ! SF measurement data (Wilson/Clover)
  !
    character(len=CHARLEN) :: fname_quark  = ""
    character(len=CHARLEN) :: fname_gsf    = "Sf_coupling_gauge."
    character(len=CHARLEN) :: fname_gsfq   = "Sf_coupling_quark."
    character(len=CHARLEN) :: fname_pcac   = "Sf_PCAC."
    character(len=CHARLEN) :: fname_ZP     = "Sf_ZP."
    character(len=CHARLEN) :: fname_ZV     = "Sf_ZV."
    character(len=CHARLEN) :: fname_ZA     = "Sf_ZA."
    character(len=CHARLEN) :: fname_pcac_solver = "Sf_PCAC_solver_log."
    character(len=CHARLEN) :: fname_gsfq_solver = "Sf_gsfq_solver_log."
    type(logfile), pointer :: logfile_gsf    => NULL()
    type(logfile), pointer :: logfile_gsfq   => NULL()
    type(logfile), pointer :: logfile_pcac   => NULL()
    type(logfile), pointer :: logfile_ZP     => NULL()
    type(logfile), pointer :: logfile_ZV     => NULL()
    type(logfile), pointer :: logfile_ZA     => NULL()
    type(logfile), pointer :: logfile_pcac_solver => NULL()
    type(logfile), pointer :: logfile_gsfq_solver => NULL()
    type(hmc_status), pointer :: status => NULL()
    type(timer) :: etime

    integer :: sf_measure_type = SF_MEASURE_NONE

    real(DP) :: kappa   = 0.0_DP
    real(DP) ::   csw   = 0.0_DP
    real(DP) :: cttilde = 1.0_DP
    integer  :: nflavor
    real(DP) :: tolerance
    real(DP) :: stout_alpha = 0.0_DP
    integer  :: stout_nstep = 0

    type(random_parameter) :: random_gsfq

    complex(DP) :: fA0(2:NTT), fP0(2:NTT)
    complex(DP) :: fA1(2:NTT), fP1(2:NTT)
    complex(DP) :: fAA(2:NTT,2:NTT)
    complex(DP) :: fAP(2:NTT,2:NTT)
    complex(DP) :: fPA(2:NTT,2:NTT)
    complex(DP) :: fPP(2:NTT,2:NTT)
    complex(DP) :: fAAc(2:NTT,2:NTT)
    complex(DP) :: fAPc(2:NTT,2:NTT)
    complex(DP) :: fPAc(2:NTT,2:NTT)
    complex(DP) :: fPPc(2:NTT,2:NTT)
    real(DP)    :: fV(2:NTT)
    real(DP)    :: f1,f1p
    complex(DP) ::  K(COL,COL,SPIN,SPIN)
    complex(DP) :: Kp(COL,COL,SPIN,SPIN)

    real(DP) :: gSF2inv
    real(DP) :: vSF

    real(DP) :: gSF2inv_q
    real(DP) :: vSF_q

    character(len=CHARLEN) :: comment(10)

  end type

  interface new
    module procedure new_sf_measurement
  end interface

  interface delete
    module procedure delete_sf_measurement
  end interface

  interface measure
    module procedure measure_sf
  end interface

  interface set_z2_noise
    module procedure set_z2_noise_fq_sf
    module procedure set_z2_noise_fq_sf_eo
  end interface

  interface set_coordinate_basis
    module procedure set_coordinate_basis_fq_sf
    module procedure set_coordinate_basis_fq_sf_eo
  end interface

  interface set_sf_classical_background
    module procedure set_sf_classical_bg
    module procedure set_sf_classical_bg_eo
  end interface

contains

subroutine read_gsfq_random(this)
!
! Initialize or restore Random Number generator
!
  use file_tools_class
  use print_status_class
  implicit none
  type(random_parameter), intent(inout) :: this
  character(len=CHARLEN) :: fname,char
  integer :: iout

  fname=fname_cont(this%run_number,this%seed_fpath,this%seed_fname)

  if (0 == nodeid) then
    write(*,'(80("="))')
    write(*,'(" SF coupling quark random number generator")')
    write(*,'("      Seed Save path :",A)')TRIM(this%seed_fpath)
    write(*,'("      Seed Save name :",A)')TRIM(this%seed_fname)
    write(*,'("               Run # :",I10)')   this%run_number
    write(*,'("         Random Seed :",I10)')   this%random_number_seed
    write(*,'(" # of Noise Estimate :",I10)')   this%nnoise
    write(*,'(80("="))')
  endif

  select case (this%run_number)
  case (0)
    call new(m_rand,0,this%random_number_seed,NPU,nodeid)
    if (nodeid == 0) then
      write(*,'(1X,"Random Number Generator Initialized")')
      write(*,'(80("-"))')
    endif
  case (1:)
    call new(m_rand,1,this%random_number_seed,NPU,nodeid)
    iout = search_free_file_unit()
    open(iout,file=fname,status='unknown',form='unformatted',action='read')
    call read(m_rand,iout)
    close(iout)
    if (nodeid == 0) then
      write(*,'(1X,"Random Number Generator Recovered")')
      write(*,'(80("-"))')
    endif
    write(char,'(1X,A," : read OK ",I3)')TRIM(fname),nodeid
    call print_status(char)
  end select
  this%run_number=this%run_number+1
  m_is_initialized = .TRUE.

  return
end subroutine

subroutine save_gsfq_random(this)
!
! Save Random Seed
!
  use file_tools_class
  implicit none
  type(random_parameter), intent(inout) :: this
  character(len=CHARLEN) :: fname
  integer :: iout

  fname = fname_cont(this%run_number,this%seed_fpath,this%seed_fname)
  iout = search_free_file_unit()
  open(iout,file=fname,status='unknown',form='unformatted',action='write')
  call save(m_rand,iout)
  close(iout)

  return
end subroutine

subroutine new_sf_measurement(this,quark_fname,status)
!
! Initialize SF measurement
!
!        this : SF measurement
! quark_fname : SF PCAC quark mass parameter file name
!      status : HMC job status
!
  use comlib, only : comlib_bcast
  use tokenize
  use file_tools_class
  use field_fermion_class, only : GPBC_THETA
  implicit none
  type(sf_measurement),     intent(inout) :: this
  character(len=*),         intent(in)    :: quark_fname
  type(hmc_status), target, intent(in)    :: status
  character(len=CHARLEN) :: str, part
  type(tokenizer) :: token
  integer :: iout,itraj,token_len

  call new(this%etime)
  call tic(this%etime)

  !---------------------------------
  ! set spatial boundary condition
  !---------------------------------
  m_theta(:) = 0.0_DP
#if    _FERMI_BOUNDARY_TYPE_Z ==  PBC
  m_theta(3) = 0.0_DP
#elif  _FERMI_BOUNDARY_TYPE_Z == APBC
  m_theta(3) = PI
#elif  _FERMI_BOUNDARY_TYPE_Z == GPBC
  m_theta(3) = GPBC_THETA
#endif

#if    _FERMI_BOUNDARY_TYPE_Y ==  PBC
  m_theta(2) = 0.0_DP
#elif  _FERMI_BOUNDARY_TYPE_Y == APBC
  m_theta(2) = PI
#elif  _FERMI_BOUNDARY_TYPE_Y == GPBC
  m_theta(2) = GPBC_THETA
#endif

#if    _FERMI_BOUNDARY_TYPE_X ==  PBC
  m_theta(1) = 0.0_DP
#elif  _FERMI_BOUNDARY_TYPE_X == APBC
  m_theta(1) = PI
#elif  _FERMI_BOUNDARY_TYPE_X == GPBC
  m_theta(1) = GPBC_THETA
#endif

  call set_tokenizer(token,' ',' ',' ')

  this%fname_quark = TRIM(quark_fname)
  this%status => status

  if (0 == nodeid) then
    iout = search_free_file_unit()
    open(iout,file=TRIM(this%fname_quark),status='old',form='formatted',action='read')
    read(iout,'(A)')this%comment(1)
    read(iout,*)    this%sf_measure_type
    read(iout,'(A)')this%comment(2)
#ifdef _CLOVER
    read(iout,*)    this%kappa,this%csw,this%cttilde
#else
    this%csw = ZERO
    this%cttilde = 1.0_DP
    read(iout,*)    this%kappa
#endif
    read(iout,*)    this%nflavor
    read(iout,*)    this%stout_alpha,this%stout_nstep
    read(iout,'(A)')this%comment(3)
    read(iout,*)    this%tolerance
    read(iout,'(A)')this%comment(4)
    read(iout,*)    this%random_gsfq%nnoise
    read(iout,'(A)')this%random_gsfq%seed_fpath
    read(iout,'(A)')this%random_gsfq%seed_fname
    read(iout,*)    this%random_gsfq%run_number
    read(iout,*)    this%random_gsfq%random_number_seed
    close(iout)
  endif

#ifndef _singlePU  
  call comlib_bcast(this%sf_measure_type,0)

  call comlib_bcast(this%kappa,0)
  call comlib_bcast(this%csw,0)
  call comlib_bcast(this%cttilde,0)

  call comlib_bcast(this%nflavor,0)
  call comlib_bcast(this%stout_alpha,0)
  call comlib_bcast(this%stout_nstep,0)

  call comlib_bcast(this%tolerance,0)

  call comlib_bcast(this%random_gsfq%nnoise,0)
  call comlib_bcast(this%random_gsfq%seed_fpath,0)
  call comlib_bcast(this%random_gsfq%seed_fname,0)
  call comlib_bcast(this%random_gsfq%run_number,0)
  call comlib_bcast(this%random_gsfq%random_number_seed,0)
#endif

  allocate(this%logfile_gsf)
  allocate(this%logfile_gsfq)
  allocate(this%logfile_pcac)
  allocate(this%logfile_ZP)
  allocate(this%logfile_ZV)
  allocate(this%logfile_ZA)
  allocate(this%logfile_pcac_solver)
  allocate(this%logfile_gsfq_solver)

  itraj = get_trajectory_number(this%status)

  !-------------------------------------------------
  ! print parameters
  !-------------------------------------------------
  if (0==nodeid) then
    write(*,'(80("="))')
    write(*,'("     SF measuement :")')
  endif


  if ( IAND(this%sf_measure_type,SF_MEASURE_GSF_GLUON) == SF_MEASURE_GSF_GLUON ) then
    if (0==nodeid) then
      write(*,'("                   : SF coupling gluon part")')
    endif

    !
    ! open SF coupling gauge part log file
    !
    write(str,'(A,I6.6)')TRIM(ADJUSTL(this%fname_gsf)),itraj
    call new(this%logfile_gsf,TRIM(str))
    call print(this%logfile_gsf,"# SF scheme coupling gluon contribution")
    write(str,'("# traj",8X,"1/(g_SF^2)",16X,"v-term",16X,"k/(g_SF^2)",15X,"k*v-term",20X,"k")')
    call print(this%logfile_gsf,TRIM(str))

  endif

  if ( IAND(this%sf_measure_type,SF_MEASURE_GSF_QUARK) == SF_MEASURE_GSF_QUARK ) then
    if (0==nodeid) then
      write(*,'("                   : SF coupling quark part")')
    endif

    call read_gsfq_random(this%random_gsfq)

    !
    ! open SF coupling quark part log file
    !
    write(str,'(A,I6.6)')TRIM(ADJUSTL(this%fname_gsfq)),itraj
    call new(this%logfile_gsfq,TRIM(str))
    call print(this%logfile_gsfq,"# SF scheme coupling quark contribution")
    write(str,'("# traj",8X,"1/(g_SF^2)",16X,"v-term",16X,"k/(g_SF^2)",15X,"k*v-term",20X,"k")')
    call print(this%logfile_gsfq,TRIM(str))

    !
    !  open SF coupling quark part solver log file
    !
    write(str,'(A,I6.6)')TRIM(ADJUSTL(this%fname_gsfq_solver)),itraj
    call new(this%logfile_gsfq_solver,TRIM(str))
    call print(this%logfile_gsfq_solver,"# SF gsf quark part quark solver log")

  endif

  if ((IAND(this%sf_measure_type,SF_MEASURE_PCAC) == SF_MEASURE_PCAC) .OR. &
 &    (IAND(this%sf_measure_type,SF_MEASURE_ZFACTOR) == SF_MEASURE_ZFACTOR)) then

    if (0==nodeid) then
      write(*,'("                   : SF PCAC correlation functions")')
    endif

    !
    ! open SF PCAC log file
    !
    write(str,'(A,I6.6)')TRIM(ADJUSTL(this%fname_pcac)),itraj
    call new(this%logfile_pcac,TRIM(str))
    call print(this%logfile_pcac,"# SF PCAC correlation functions")
    write(str,'("# traj itt",9X,"fA(itt)",17X,"fP(itt)",17X,"fA''(itt)",16X,"fP''(itt)")')
    call print(this%logfile_pcac,TRIM(str))

    !
    ! open SF PCAC quark solver log file
    !
    write(str,'(A,I6.6)')TRIM(ADJUSTL(this%fname_pcac_solver)),itraj
    call new(this%logfile_pcac_solver,TRIM(str))
    call print(this%logfile_pcac_solver,"# SF PCAC correlation solver log")

  endif

  if ( IAND(this%sf_measure_type,SF_MEASURE_ZFACTOR) == SF_MEASURE_ZFACTOR ) then
    if (0==nodeid) then
      write(*,'("                   : SF Z factor correlation functions")')
    endif

    !
    ! open SF ZP,ZV,ZA log file
    !
    write(str,'(A,I6.6)')TRIM(ADJUSTL(this%fname_ZP)),itraj
    call new(this%logfile_ZP,TRIM(str))
    call print(this%logfile_ZP,"# SF ZP correlation functions f1, f1 is time independent constant.")
    write(str,'("# traj itt",12X,"f1",21X,"f1p")')
    call print(this%logfile_ZP,TRIM(str))

    write(str,'(A,I6.6)')TRIM(ADJUSTL(this%fname_ZV)),itraj
    call new(this%logfile_ZV,TRIM(str))
    call print(this%logfile_ZV,"# SF ZV correlation functions fV(t)")
    write(str,'("# traj itt",9X,"fV(itt)")')
    call print(this%logfile_ZV,TRIM(str))

    write(str,'(A,I6.6)')TRIM(ADJUSTL(this%fname_ZA)),itraj
    call new(this%logfile_ZA,TRIM(str))
    call print(this%logfile_ZA,"# SF ZA correlation functions fAA(t,s), fAP(t,s), fPA(t,s), fPP(t,s)")

  endif

  if (0==nodeid) then
    write(*,'("        Trajectory :",I10)')  itraj
    if (0 < this%stout_nstep) then
      write(*,'(80("-"))')
      write(*,'(" Stout semared fermion :")')
      write(*,'("             alpha :",F16.9)')this%stout_alpha
      write(*,'("             nstep :",I3   )')this%stout_nstep
    endif
    write(*,'("     PCAC:   Kappa :",F16.9)')this%kappa
#ifdef _CLOVER
    write(*,'("     PCAC:     Csw :",F16.9)')this%csw
    write(*,'("     PCAC: Cttilde :",F16.9)')this%cttilde
#endif
    write(*,'("   SF coupling: Nf :",I3)')   this%nflavor
    write(*,'("  Solver tolerance :",E24.15)')this%tolerance
    write(*,'(80("-"))')
  endif

  return
end subroutine

subroutine delete_sf_measurement(this)
!
! Delete SF measurement
!
  use file_tools_class
  implicit none
  type(sf_measurement),  intent(inout) :: this
  real(DP) :: etime
  integer :: iout

  if ( IAND(this%sf_measure_type,SF_MEASURE_GSF_QUARK) == SF_MEASURE_GSF_QUARK ) then
    call save_gsfq_random(this%random_gsfq)
  endif

  !-------------------------------
  ! save status
  !-------------------------------
  if (0==nodeid) then
    iout = search_free_file_unit()
    open(iout,file=TRIM(this%fname_quark),status='old',form='formatted',action='write')
    write(iout,'(A)') TRIM(ADJUSTL(this%comment(1)))
    write(iout,'(I3)')             this%sf_measure_type
    write(iout,'(A)') TRIM(ADJUSTL(this%comment(2)))
#ifdef _CLOVER
    write(iout,'(E24.15,2ES24.14)')this%kappa,this%csw,this%cttilde
#else
    write(iout,'(E24.15)')         this%kappa
#endif
    write(iout,'(I3)')             this%nflavor
    write(iout,'(E24.15,I3)')      this%stout_alpha,this%stout_nstep
    write(iout,'(A)') TRIM(ADJUSTL(this%comment(3)))
    write(iout,'(E24.15)')         this%tolerance
    write(iout,'(A)') TRIM(ADJUSTL(this%comment(4)))
    write(iout,'(I4)')             this%random_gsfq%nnoise
    write(iout,'(A)') TRIM(ADJUSTL(this%random_gsfq%seed_fpath))
    write(iout,'(A)') TRIM(ADJUSTL(this%random_gsfq%seed_fname))
    write(iout,'(I10)')            this%random_gsfq%run_number
    write(iout,'(I14)')            this%random_gsfq%random_number_seed
    close(iout)
  endif

  this%status => NULL()

  if (IAND(this%sf_measure_type,SF_MEASURE_GSF_GLUON) == SF_MEASURE_GSF_GLUON) then
    call delete(this%logfile_gsf)
  endif

  if (IAND(this%sf_measure_type,SF_MEASURE_GSF_QUARK) == SF_MEASURE_GSF_QUARK) then
    call delete(this%logfile_gsfq)
    call delete(this%logfile_gsfq_solver) 
  endif

  if ((IAND(this%sf_measure_type,SF_MEASURE_PCAC) == SF_MEASURE_PCAC) .OR. &
 &    (IAND(this%sf_measure_type,SF_MEASURE_ZFACTOR) == SF_MEASURE_ZFACTOR)) then
    call delete(this%logfile_pcac)
    call delete(this%logfile_pcac_solver) 
  endif

  if (IAND(this%sf_measure_type,SF_MEASURE_ZFACTOR) == SF_MEASURE_ZFACTOR) then
    call delete(this%logfile_ZP)
    call delete(this%logfile_ZV)
    call delete(this%logfile_ZA)
  endif

  deallocate(this%logfile_gsf)
  deallocate(this%logfile_gsfq)
  deallocate(this%logfile_pcac)
  deallocate(this%logfile_ZP)
  deallocate(this%logfile_ZV)
  deallocate(this%logfile_ZA)
  deallocate(this%logfile_pcac_solver)
  deallocate(this%logfile_gsfq_solver)

  this%logfile_gsf  => NULL()
  this%logfile_gsfq => NULL()
  this%logfile_pcac => NULL()
  this%logfile_ZP   => NULL()
  this%logfile_ZV   => NULL()
  this%logfile_ZA   => NULL()
  this%logfile_pcac_solver => NULL()
  this%logfile_gsfq_solver => NULL()

  call toc(this%etime)
  etime = get_elapse(this%etime)

  if (0==nodeid) then
    write(*,'(80("-"))')
    write(*,'(" SF measurement statistics")')
    write(*,'("          time (sec) :",F12.4)') etime
    write(*,'(80("="))')
  endif

  return
end subroutine

subroutine measure_sf(this,gauge,u)
!
! Measure SF scheme coupling and PCAC correlation functions
! and correlation functions for ZP, ZV, ZA factors
!
!   this : SF measurement
!  gauge : gauge action couplings (SF boundary couplings are used)
!      u : gauge link
!
  use hmc_gluon_class, only : gluon_rg_sf
  implicit none
  type(sf_measurement),  intent(inout) :: this
  type(gluon_rg_sf),     intent(in) :: gauge
  type(vfield_gluon_wg), intent(inout), target :: u

  type(quark_prop), allocatable :: qp
  type(qprop_params), target :: params
  type(src_wavefunc_obj), target, allocatable :: wf
#ifdef _STOUT
  type(stout_smearing_parameters), target :: stout
#endif
  type(vfield_gluon_wg), pointer :: utmp

  utmp => NULL()
!======
!  call set_sf_boundary(u)
! for check
!  call set_identity(u)
!======

  !=========================================
  ! compute SF scheme couplings gluon part
  !=========================================
  if ((IAND(this%sf_measure_type,SF_MEASURE_GSF_GLUON) == SF_MEASURE_GSF_GLUON) .OR. &
 &    (IAND(this%sf_measure_type,SF_MEASURE_GSF_QUARK) == SF_MEASURE_GSF_QUARK) )  then

    call get_sf_scheme_coupling(this,gauge,u)

  endif

  !=========================================
  ! compute SF scheme couplings quark part
  !=========================================
  if ( IAND(this%sf_measure_type,SF_MEASURE_GSF_QUARK) == SF_MEASURE_GSF_QUARK )  then

#ifdef _GGGXXX_
    call new(params)
#ifdef _CLOVER
    allocate(quark_clover::params%mass)
#else
    allocate(quark_wilson::params%mass)
#endif
    select type(p => params%mass)
#ifdef _CLOVER
    type is (quark_clover)
#else
    type is (quark_wilson)
#endif
      call new(p,id=0)
      call set_kappa(p,this%kappa)
#ifdef _CLOVER
      call set_csw(p,this%csw)
      call set_cttilde(p,this%cttilde)
#endif
    end select

    params%tol        =  this%tolerance
    params%solver_log => this%logfile_gsfq_solver
#endif

    call get_sf_scheme_coupling_quark(this,gauge,u)

  endif

  !================================================
  ! compute SF PCAC/Z-factor corelation functions
  !================================================
  if ((IAND(this%sf_measure_type,SF_MEASURE_PCAC)      == SF_MEASURE_PCAC)      .OR. &
 &    (IAND(this%sf_measure_type,SF_MEASURE_ZFACTOR)   == SF_MEASURE_ZFACTOR) ) then

    call new(params)

#ifdef _CLOVER
    allocate(quark_clover::params%mass)
#else
    allocate(quark_wilson::params%mass)
#endif
    select type(p => params%mass)
#ifdef _CLOVER
    type is (quark_clover)
#else
    type is (quark_wilson)
#endif
      call new(p,id=0)
      call set_kappa(p,this%kappa)
#ifdef _CLOVER
      call set_csw(p,this%csw)
      call set_cttilde(p,this%cttilde)
#endif
    end select

    params%tol        =  this%tolerance
    params%solver_log => this%logfile_pcac_solver

    allocate(qp,wf)
    call new(qp)
    call new(wf)
    qp%parameter     => params
    qp%wavefunc_sous => wf
    qp%smear_sous    = 1
    qp%wavefunc_sink => NULL()
    qp%smear_sink    = 0

#ifdef _STOUT
    !-----------------
    ! stout smearing
    !-----------------
    if (0 < this%stout_nstep) then
      call new(stout,this%stout_alpha,this%stout_nstep,type=STOUT_TYPE_ALL,is_smear_only=.TRUE.)
      call print(stout)
      call smear(stout,u)
      utmp => stout%usmr(1)   ! When smearing only is selected, the smeard link is stored in usmr(1)
    else
      utmp => u
    endif
#else
    utmp => u
#endif

    call set_sf_source(this,qp,utmp)
    call get_prop(qp,utmp)
    call change_normalization(this,qp)

    if ((IAND(this%sf_measure_type,SF_MEASURE_PCAC)      == SF_MEASURE_PCAC)      .OR. &
 &      (IAND(this%sf_measure_type,SF_MEASURE_ZFACTOR)   == SF_MEASURE_ZFACTOR) ) then
      call get_pcac(this,qp)
    endif

    if (IAND(this%sf_measure_type,SF_MEASURE_ZFACTOR) == SF_MEASURE_ZFACTOR) then
      call get_ZP(this,qp,utmp)
      call get_ZV(this,qp)
      call get_ZA(this,qp,utmp)
    endif

#ifdef _STOUT
    if (0 < this%stout_nstep) then
      call delete(stout)
    endif
#endif

    call delete(qp)
    call delete(wf)
    call delete(params)
    deallocate(params%mass)
    deallocate(qp,wf)

  endif

  return
end subroutine


subroutine set_sf_source(this,qp,u)
!
! set SF PCAC correlation function boundary source fields on a quark propagator
!
! itt=2   (t=1)    qp(ic,jc,is=js,js=1:2,t,z,y,x) = tilde{c}_t (U_4(t-1,z,y,x)^+)_{ic,jc} GFactor(z,y,x)
! itt=NTT (t=T-1)  qp(ic,jc,is=js,js=3:4,t,z,y,x) = tilde{c}_t (U_4(t  ,z,y,x)  )_{ic,jc} GFactor(z,y,x)
! (t = itt-1)
!
!    qp : quark propagator
!     u : gauge link field
!
! GFactor(z,y,x) = Exp( I* ((theta_x/N_x)*x + (theta_y/N_y)*y + (theta_z/N_z)*z )
!
! GFactor(z,y,x) is required for generic spatial boundary condition
! as the boundary operator is not gauge invariant.
! 
! The original paper by Luescher et al. uses priodic boundary condition instead of generic boundary condition
! by applying Abelian-guage transformation.
! The hopping matrix contains extra U(1) phase facotr in the periodic boundary condition form.
! The original paper uses this gauge transformed form in the  perturbative analysis and correlation 
! function definitions.
!
! This program set (NewHMC) uses generic boundary condition for which we impose the condition at the
! ghost boundary copy routine.  The difference between the original one and my program appears in 
! the boundary operator up to the gauge transformation.
!
! NewHMC definition:
!
!  psi(n+\vec{L}) = exp( I*(theta_x + theta_y + theta_z) ) psi(n)   
!
! Luescher's definition:
!
!  psi'(n+\vec{L}) = psi'(n)   
!
! Gauge transformation:
!
!  psi'(n) = exp(-I*(\vec{theta/L}.\vec{n})) psi(n) = Omega(\vec{n)) psi(n)
!  ___                                       ___                         ___
!  psi'(n) = exp( I*(\vec{theta/L}.\vec{n})) psi(n) = Omega(\vec{n))^dag psi(n)
!
! We uses
!
!  Gfactor(z,y,x) = Omega(\vec(n))^dag
!
  use error_class
  implicit none
  type(sf_measurement),  intent(in)    :: this
  type(quark_prop),      intent(inout) :: qp
  type(vfield_gluon_wg), intent(in)    :: u
  integer :: ieo,ix,iy,iz,ieoxyz,ith,ic,jc,is,js,isp,itt
  real(DP) :: arg, cttilde
  integer :: itx,ity,itz
  complex(DP) :: gfac

  cttilde = this%cttilde

  if (allocated(qp%prop)) deallocate(qp%prop)
  allocate(qp%prop(COL,COL,SPIN,SPIN,NZ*NY*NX,NTT))

  call clear_prop(qp)

  do ieo=0,1
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,isp,itt,js,is,jc,ic,itx,ity,itz,arg,gfac)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+iz+iy+ix+ieo,2)
    isp = ispace(ix,iy,iz)

!=======================================
! Set itt = 2 boundary (t = 1)
!
!     ith = 1-ieoxyz,..,NTH-ieoxyz
!     itt = 2*ith + ieoxyz
!     itt = 2 => ith = 1, ieoxyz = 0  (t = 1)
!     itt = 1 => ith = 0, ieoxyz = 1  (t = 0)
!
!  Dirac rep.  (1+gamma_4)/2 projection
! set spin = 1,2 (upper 2 spinor) only.
!
!=======================================

    itx = ix + ipsite(1)*NX
    ity = iy + ipsite(2)*NY
    itz = iz + ipsite(3)*NZ
    arg = (itx-1)*m_theta(1)/NTX  &
 &       +(ity-1)*m_theta(2)/NTY  &
 &       +(itz-1)*m_theta(3)/NTZ 
    gfac = CMPLX(COS(arg),SIN(arg),kind=DP)
    if (1==ieoxyz) then
      itt = 2
      do is=1,2    !<== upper 2 spinor
      do jc=1,COL
      do ic=1,COL
        qp%prop(ic,jc,is,is,isp,itt) = conjg(u%eo(ieo)%mu(4)%s(0,iz,iy,ix)%u(jc,ic))*gfac*cttilde  ! <== (U_4(itt=1))^+_(ic,jc)
      enddo
      enddo
      enddo
    endif

!=======================================
! Set itt = NTT boundary (t = T-1)
!
!     ith = 1-ieoxyz,..,NTH-ieoxyz
!     itt = 2*ith + ieoxyz
!     itt = NTT => ith = NTH, ieoxyz = 0 
!
!  Dirac rep.  (1-gamma_4)/2 projection
! set spin = 3,4 (lower 2 spinor) only.
!
!=======================================
    if (0==ieoxyz) then
      itt = NTT
      do is=3,4    !<== lower 2 spinor
      do jc=1,COL
      do ic=1,COL
        qp%prop(ic,jc,is,is,isp,itt) = u%eo(ieo)%mu(4)%s(NTH,iz,iy,ix)%u(ic,jc)*gfac*cttilde   ! <== U_4(itt=NTT)_(ic,jc)
      enddo
      enddo
      enddo
    endif

  enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine set_sf_source_connected(qp,qp_src,t_src,channel)
!
! set sf source for connected diagrams
!
! qp(\vec{x},t;s) = Sum_{\vec{y}} S(\vec{x},t;\vec{y},s) Gamma_T q_src(\vec{y},s)
!
! Gamma_T : gamma4*gamma5 or gamma5
!       s :  t_src
!
!  qp     : connected propagator
!  qp_src : t=1 boundary sourced propagetor H(x,t) for spin_src=1,2, H'(x,t) for spin_src=3,4
!   t_src : connection time (s)
! channel : Channel 'A' or 'P'
!
  use error_class
  use field_fermion_class, only : igamma, zgamma
  implicit none
  type(quark_prop),      intent(inout) :: qp
  type(quark_prop),      intent(in)    :: qp_src
  integer,               intent(in)    :: t_src
  character(len=1),      intent(in)    :: channel
  character(len=CHARLEN) :: str
  integer :: isp,itt,ic,jc,is,js
  integer :: igmY(SPIN)
  complex(DP) :: zgmY(SPIN)

  select case(channel)
  case('A')
    !--------------------------------------
    ! Gamma_Y = Gamma_A = gamma_4 gamma_5
    !--------------------------------------
    do is=1,SPIN
      igmY(is) = igamma(igamma(is,4),5)
      zgmY(is) = zgamma(igamma(is,4),5)*zgamma(is,4)
    enddo

  case('P')
    !--------------------------------------
    ! Gamma_Y = Gamma_P = gamma_5
    !--------------------------------------
    do is=1,SPIN
      igmY(is) = igamma(is,5)
      zgmY(is) = zgamma(is,5)
    enddo

  case default

    write(str,'("set_sf_source_connecetd: channel must be ""A"" or ""P"".")')
    call error_stop(str)

  end select

  if (0 == nodeid) then
    write(*,'(80("-"))')
    write(*,'("set_sf_source_connecetd: ",A," channnel connected at t=",I3)')channel,t_src
  endif

  if (allocated(qp%prop)) deallocate(qp%prop)
  allocate(qp%prop(COL,COL,SPIN,SPIN,NZ*NY*NX,NTT))

  call clear_prop(qp)

  !------------------------------------------------------
  ! set source at t=t_src
  !------------------------------------------------------
  itt = t_src
!$OMP PARALLEL DO PRIVATE(isp,is,js,ic,jc)
  do isp = 1,NSPACE
    do js=1,SPIN
    do is=1,SPIN
      do jc=1,COL
      do ic=1,COL
        qp%prop(ic,jc,is,js,isp,itt) = zgmY(is)*qp_src%prop(ic,jc,igmY(is),js,isp,itt)
      enddo
      enddo
    enddo
    enddo
  enddo

  return
end subroutine

subroutine change_normalization(this,qp)
! 
! change field normalization from kappa form to mass form
!
! prop <=   (2*kappa) * prop
!
! bare mass = 1/(2*kappa) - 4
!
  type(sf_measurement), intent(in) :: this
  type(quark_prop),     intent(inout) :: qp
  real(DP) :: kappa,mass
  integer :: ic,jc,is,js,isp,itt

  kappa = this%kappa
  mass = 1.0_DP/(2*kappa) - 4.0_DP
  if (0 == nodeid) write(*,'("% kappa =",E24.15," mass =",E24.15)')kappa,mass

  do itt = 1,NTT
!$OMP PARALLEL DO PRIVATE(ic,jc,is,js,isp)
  do isp = 1,NSPACE
    do js=1,SPIN
    do is=1,SPIN
    do jc=1,COL
    do ic=1,COL
      qp%prop(ic,jc,is,js,isp,itt) = qp%prop(ic,jc,is,js,isp,itt)*(2.0_DP*kappa)
    enddo
    enddo
    enddo
    enddo
  enddo
  enddo

  return
end subroutine
  
subroutine get_pcac(this,qp)
!
! measure SF PCAC correlation functions.
!
!   fA(x_0)   = 1/2/Vol Sum_{vec{x}} Tr[qp^+(vec{x},x_0) [-gamma_0] qp(vec{x},x_0)]
!   fP(x_0)   = 1/2/Vol Sum_{vec{x}} Tr[qp^+(vec{x},x_0)            qp(vec{x},x_0)]
!   use qp(:,:,:,js=1,2) only (Dirac Rep.(1+gamma4)/2)
! 
!   fA'(T-x_0)= 1/2/Vol Sum_{vec{x}} Tr[qp^+(vec{x},x_0) [+gamma_0] qp(vec{x},x_0)]
!   fP'(T-x_0)= 1/2/Vol Sum_{vec{x}} Tr[qp^+(vec{x},x_0)            qp(vec{x},x_0)]
!   use qp(:,:,:,js=3,4) only (Dirac Rep.(1-gamma4)/2)
! 
!     this : sf measuremente 
!       qp : quark propagator
! 
  use comlib, only : comlib_sumcast
  use field_fermion_class, only : igamma, zgamma
  implicit none
  type(sf_measurement), intent(inout) :: this
  type(quark_prop),     intent(in) :: qp
  integer :: itt,isp,ic,is,jc,js
  integer :: isnk
  complex(DP) :: zgm
  complex(DP) :: ztmp0_A,ztmp1_A
  complex(DP) :: ztmp0_P,ztmp1_P
  integer :: itraj
  character(len=CHARLEN) :: str

  this%fA0(:)= Z0
  this%fP0(:)= Z0
  this%fA1(:)= Z0
  this%fP1(:)= Z0

  do itt=2,NTT
  ztmp0_A = Z0; ztmp0_P = Z0
  ztmp1_A = Z0; ztmp1_P = Z0
!$OMP PARALLEL DO PRIVATE(isp,js,is,jc,ic,isnk,zgm) REDUCTION(+:ztmp0_A,ztmp0_P,ztmp1_A,ztmp1_P)
  do isp = 1,NSPACE
!********* fA
    do js=1,SPIN/2
    do is=1,SPIN
      isnk = igamma(is,4)
      zgm  = zgamma(is,4)
      do jc=1,COL
      do ic=1,COL
        ztmp0_A = ztmp0_A + zgm*qp%prop(ic,jc,isnk,js  ,isp,itt)   &
 &                       *conjg(qp%prop(ic,jc,is  ,js  ,isp,itt))
        ztmp1_A = ztmp1_A + zgm*qp%prop(ic,jc,isnk,js+2,isp,itt)   &
 &                       *conjg(qp%prop(ic,jc,is  ,js+2,isp,itt))
      enddo
      enddo
    enddo
    enddo
!********* fP
    do js=1,SPIN/2
    do is=1,SPIN
      do jc=1,COL
      do ic=1,COL
        ztmp0_P = ztmp0_P +     qp%prop(ic,jc,is,js  ,isp,itt)  &
     &                   *conjg(qp%prop(ic,jc,is,js  ,isp,itt))
        ztmp1_P = ztmp1_P +     qp%prop(ic,jc,is,js+2,isp,itt)  &
     &                   *conjg(qp%prop(ic,jc,is,js+2,isp,itt))
      enddo
      enddo
    enddo
    enddo
  enddo
  this%fA0(itt)       = ztmp0_A
  this%fP0(itt)       = ztmp0_P
  this%fA1(NTT+2-itt) = ztmp1_A
  this%fP1(NTT+2-itt) = ztmp1_P
  enddo

#ifndef _singlePU
  call comlib_sumcast(this%fA0(:))
  call comlib_sumcast(this%fP0(:))
  call comlib_sumcast(this%fA1(:))
  call comlib_sumcast(this%fP1(:))
#endif

  do itt=2,NTT
    this%fA0(itt)=-this%fA0(itt)/REAL(2*NTX*NTY*NTZ,kind=DP)
    this%fP0(itt)= this%fP0(itt)/REAL(2*NTX*NTY*NTZ,kind=DP)
    this%fA1(itt)= this%fA1(itt)/REAL(2*NTX*NTY*NTZ,kind=DP)
    this%fP1(itt)= this%fP1(itt)/REAL(2*NTX*NTY*NTZ,kind=DP)
  enddo

  itraj = get_trajectory_number(this%status)

  if (0==nodeid) then
    do itt=2,NTT
      write(str,'(I6,I4,4E24.16)')itraj,itt, &
 &    REAL(this%fA0(itt),kind=DP),REAL(this%fP0(itt),kind=DP), &
 &    REAL(this%fA1(itt),kind=DP),REAL(this%fP1(itt),kind=DP)
      call print(this%logfile_pcac,TRIM(str))
    enddo
  endif

  return
end subroutine

subroutine get_ZP(this,qp,u)
!
! Calculate SF scheme ZP factor
!
!  qp contains  
!
!  qp(ic,jc,is,js=1:2,y) = H(y)_{ic,jc,is,js=1:2}
!  qp(ic,jc,is,js=3:4,y) = H'(y)_{ic,jc,is,js=3:4}
!
!    H(y) = \sum_{\vec{x}}[\psi(y)\var{zeta(\vec{x})}]
!   H'(y) = \sum_{\vec{x}}[\psi(y)\var{zeta'(\vec{x})}]
!
! For ZP, we need K or Kp defined by
!
!  K  = (1/NV) \sum_{\vec{x},\vec{y}}[\'zeta(\vec{x})\var{zeta(\vec{y})}],
!
!  Kp = (1/NV) \sum_{\vec{x},\vec{y}}[\zeta(\vec{x})\var{zeta'(\vec{y})}],
!
!  (NV = NTX*NTY*NTZ),
!
! This is from boundary(t=0) to boundary(t=T) propagator.
!
! K  = (\tilde{c}_t/NV) * \sum_{\vec{x}} gfac(\vec(x)) P_4^+ U^+_4(\vec{x},T-1) H(\vec{x},T-1)
!
! Kp = (\tilde{c}_t/NV) * \sum_{\vec{x}} gfac(\vec(x)) P_4^-   U_4(\vec{x},  0) H'(\vec{x}, 1)
! 
! gfac(\vec{x}) = exp(-I*(\vec{\theta/L}.\vec{x}))
!
! Thus K only has the following spin components;  K(ic,jc,is=1:2,js=1:2)
!     Kp only has the following spin components; Kp(ic,jc,is=3:4,js=3:4)
!
! The relation
!
!    gamma_5 Kp^dag gamma_5 = K  
!
!  must holds.
!
!  ZP is constructed from fP(t) and f1 correlators.
!
! f1 = (1/2) Tr[K^dag K] = (1/2) sum_{ic=1:3,jc=1:3,is=1:2,js=1:2} abs(K(ic,jc,is,js))**2
!
  use comlib, only : comlib_sumcast
  type(sf_measurement),  intent(inout) :: this
  type(quark_prop),      intent(in) :: qp
  type(vfield_gluon_wg), intent(in) :: u
  complex(DP) ::  K(COL,COL,SPIN,SPIN)
  complex(DP) :: Kp(COL,COL,SPIN,SPIN)
  complex(DP) :: ctmp(COL*COL*SPIN*SPIN)
  real(DP) :: rtmp,rtmp2
  real(DP) :: arg,cttilde
  complex(DP) :: gfac
  character(len=CHARLEN) :: str
  integer :: itx,ity,itz
  integer :: itraj
  integer :: ieo,ix,iy,iz,ieoxyz,isp,ic,jc,is,js,itt
  integer :: ith,ics

  cttilde = this%cttilde

   K(:,:,:,:) = Z0
  Kp(:,:,:,:) = Z0

  do ieo=0,1
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,isp,itt,ith,ic,jc,is,js,itx,ity,itz,arg,gfac) REDUCTION(+:K,Kp)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+iz+iy+ix+ieo,2)
    isp = ispace(ix,iy,iz)    

    itx = ix + ipsite(1)*NX
    ity = iy + ipsite(2)*NY
    itz = iz + ipsite(3)*NZ
    arg = (itx-1)*m_theta(1)/NTX  &
 &       +(ity-1)*m_theta(2)/NTY  &
 &       +(itz-1)*m_theta(3)/NTZ 
    gfac = CMPLX(COS(arg),-SIN(arg),kind=DP)

!====================================
! itt = NTT boundary (t = T-1)
!
!   ith = 1-ieoxyz,..,NTH-ieoxyz
!   itt = 2*ith + ieoxyz
!   itt = NTT => ith = NTH, ieoxyz = 0 
!
! this corresponds to ith = NTH, ieoxyz = 0
!
!====================================
    if (0 == ieoxyz) then
      itt = NTT
      ith = NTH
      do js=1,SPIN/2  ! upper 2 spinor (P_4^+ projected)
      do is=1,SPIN/2  ! upper 2 spinor (P_4^+ projected)
        do jc=1,COL
        do ic=1,COL

          K(ic,jc,is,js) = K(ic,jc,is,js)  &
 &                       +(  CONJG(u%eo(ieo)%mu(4)%s(ith,iz,iy,ix)%u(1,ic))  &
 &                          * qp%prop(1,jc,is,js,isp,itt)  &
 &                         + CONJG(u%eo(ieo)%mu(4)%s(ith,iz,iy,ix)%u(2,ic))  &
 &                          * qp%prop(2,jc,is,js,isp,itt)  &
 &                         + CONJG(u%eo(ieo)%mu(4)%s(ith,iz,iy,ix)%u(3,ic))  &
 &                          * qp%prop(3,jc,is,js,isp,itt) ) * gfac
  
        enddo
        enddo
      enddo
      enddo
    endif

!=======================================
! itt = 2 boundary (t = 1)
!
!     ith = 1-ieoxyz,..,NTH-ieoxyz
!     itt = 2*ith + ieoxyz
!     itt = 2 => ith = 1, ieoxyz = 0  (t = 1)
!     itt = 1 => ith = 0, ieoxyz = 1  (t = 0)
!
!=======================================
    if (1==ieoxyz) then
      itt = 2  ! (t=1)
      ith = 0  ! (t=0)
      do js=SPIN/2+1,SPIN  ! lower 2 spinor (P_4^- projected)
      do is=SPIN/2+1,SPIN  ! lower 2 spinor (P_4^- projected)
        do jc=1,COL
        do ic=1,COL

          Kp(ic,jc,is,js) = Kp(ic,jc,is,js)  &
 &                        +(  u%eo(ieo)%mu(4)%s(ith,iz,iy,ix)%u(ic,1)  &
 &                           * qp%prop(1,jc,is,js,isp,itt)  &
 &                          + u%eo(ieo)%mu(4)%s(ith,iz,iy,ix)%u(ic,2)  &
 &                           * qp%prop(2,jc,is,js,isp,itt)  &
 &                          + u%eo(ieo)%mu(4)%s(ith,iz,iy,ix)%u(ic,3)  &
 &                           * qp%prop(3,jc,is,js,isp,itt) ) * gfac

        enddo
        enddo
      enddo
      enddo
    endif

  enddo
  enddo
  enddo
  enddo

#ifndef _singlePU
  ctmp(:) = RESHAPE(K,SHAPE(ctmp))
  call comlib_sumcast(ctmp)
  K(:,:,:,:) = RESHAPE(ctmp,SHAPE(K))

  ctmp(:) = RESHAPE(Kp,SHAPE(ctmp))
  call comlib_sumcast(ctmp)
  Kp(:,:,:,:) = RESHAPE(ctmp,SHAPE(Kp))
#endif

   K(:,:,:,:) =  K(:,:,:,:)*cttilde/REAL(NTX*NTY*NTZ,kind=DP) 
  Kp(:,:,:,:) = Kp(:,:,:,:)*cttilde/REAL(NTX*NTY*NTZ,kind=DP) 
   this%K(:,:,:,:) =  K(:,:,:,:)
  this%Kp(:,:,:,:) = Kp(:,:,:,:)

  if (0 == nodeid) then

    write(*,'(80("="))')
    write(*,'("%  Check K = gamma_5 Kp^+ gamma_5")')
    write(*,'("% ic jc is js",16X,"K(ic,jc,is,js)",34X,"Kp(jc,ic,js+2,is+2)^*",34X,"diff")')
    do js=1,SPIN/2
    do is=1,SPIN/2
    do jc=1,COL
    do ic=1,COL
      write(*,'("%",4I3,6E24.16)')ic,jc,is,js,  &
 &         K(ic,jc,is,js),CONJG(Kp(jc,ic,js+SPIN/2,is+SPIN/2)), &
 &         K(ic,jc,is,js)-CONJG(Kp(jc,ic,js+SPIN/2,is+SPIN/2))
    enddo
    enddo
    enddo
    enddo
    write(*,'(80("="))')

  endif

  !----------------------------
  ! f1 = (1/2) * Tr[K^dag K]
  !----------------------------
  rtmp  = 0.0_DP
  rtmp2 = 0.0_DP
  do js=1,SPIN/2
  do is=1,SPIN/2
  do jc=1,COL
  do ic=1,COL
    rtmp  = rtmp  + REAL( this%K(ic,jc,is,js),kind=DP)**2 &
 &                +AIMAG( this%K(ic,jc,is,js))**2

  !
  ! for check f1 = (1/2) * Tr[Kp^dag Kp] 
  !
    rtmp2 = rtmp2 + REAL(this%Kp(ic,jc,is+SPIN/2,js+SPIN/2),kind=DP)**2 &
 &                +AIMAG(this%Kp(ic,jc,is+SPIN/2,js+SPIN/2))**2
  enddo
  enddo
  enddo
  enddo
  this%f1  = rtmp *0.5_DP
  this%f1p = rtmp2*0.5_DP

  !----------------------------
  ! f1 data save 
  !----------------------------
  itraj = get_trajectory_number(this%status)

  if (0 == nodeid) then
    do itt=2,NTT
      write(str,'(I6,I4,2E24.16)')itraj,itt,this%f1,this%f1p
      call print(this%logfile_ZP,TRIM(str))
    enddo
  endif

  return
end subroutine

subroutine get_ZV(this,qp)
!
! Calculate SF scheme ZV factor
!
!  qp contains  
!
!  qp(ic,jc,is,js=1:2,y) =  H(y)_{ic,jc,is,js=1:2}
!  qp(ic,jc,is,js=3:4,y) = H'(y)_{ic,jc,is,js=3:4}
!
!    H(y) = \sum_{\vec{x}}[\psi(y)\var{zeta(\vec{x})}]
!   H'(y) = \sum_{\vec{x}}[\psi(y)\var{zeta'(\vec{x})}]
!
! For ZV,
!
!  fV(t) = (1/2/NV)
!        * \sum_{\vec{x}} Re[Tr[  K^dag \gamma_5 H'(\vec{x),t)^{dag}
!                              \gamma_5 \gamma_4  H(\vec{x},t)]]
!
!  NV = (NTX*NTY*NTZ)
!
!  fV(t) = (1/2/NV)
!        * \sum_{\vec{x}} Re[Tr[
!                      \gamma_5 \gamma_4  H(\vec{x},t)
!                       K^{dag} \gamma_5 H'(\vec{x),t)^{dag} ]]
!        = (1/2/NV)
!        * \sum_{\vec{x}} Re[Tr[
!                      \gamma_5 \gamma_4  H(\vec{x},t)
!                           (H'(\vec{x),t) \gamma_5 K)^{dag} ]]
!
  use comlib, only : comlib_sumcast
  use field_fermion_class, only : igamma, zgamma
  type(sf_measurement),  intent(inout) :: this
  type(quark_prop),      intent(in) :: qp
  real(DP) :: rtmp,rtmp2
  character(len=CHARLEN) :: str
  complex(DP) :: ztmp
  complex(DP) :: zgm5gm4,zgm5,zgm4
  integer ::     igm5gm4,igm5,igm4
  integer :: itraj
  integer :: itt,isp,ic,jc,is,js,ks
  type(quark_prop), allocatable :: pp

  allocate(pp)
  call new(pp)
  if (allocated(pp%prop)) deallocate(pp%prop)
  allocate(pp%prop(COL,COL,SPIN,SPIN,NZ*NY*NX,NTT))

  !================================================
  ! pp(y) = H'(y) gamma_5 K
  !================================================
  do itt = 2,NTT
!$OMP PARALLEL DO PRIVATE(isp,js,is,jc,ic,igm5,zgm5)
  do isp = 1,NSPACE
    do js=1,SPIN/2
    do is=1,SPIN
    do jc=1,COL
    do ic=1,COL
      pp%prop(ic,jc,is,js,isp,itt) = Z0
      do ks=SPIN/2+1,SPIN
        igm5 = igamma(ks,5)
        zgm5 = zgamma(ks,5)
        pp%prop(ic,jc,is,js,isp,itt) = pp%prop(ic,jc,is,js,isp,itt)   &
 &                             + zgm5*(qp%prop(ic, 1,is,ks,isp,itt)*this%K( 1,jc,igm5,js)  &
 &                                   + qp%prop(ic, 2,is,ks,isp,itt)*this%K( 2,jc,igm5,js)  &
 &                                   + qp%prop(ic, 3,is,ks,isp,itt)*this%K( 3,jc,igm5,js) )
      enddo
    enddo
    enddo
    enddo
    enddo
  enddo
  enddo

  !==============================================================================
  ! fV(t) = (1/(2*NV))ReTr[\gamma_5 \gamma_4 H(\vec{x},t) pp(\vec{x},t)^dag ]
  !==============================================================================
  do itt = 2,NTT
  ztmp = Z0
!$OMP PARALLEL DO PRIVATE(isp,js,is,jc,ic,igm5,zgm5,igm5gm4,zgm5gm4) REDUCTION(+:ztmp)
  do isp = 1,NSPACE
    do js=1,SPIN/2
    do is=1,SPIN
      igm5    = igamma(is,5)
      zgm5    = zgamma(is,5)
      igm5gm4 = igamma(igm5,4)
      zgm5gm4 = zgamma(igm5,4)*zgm5
      do jc=1,COL
      do ic=1,COL
        ztmp = ztmp + zgm5gm4*qp%prop(ic,jc,igm5gm4,js,isp,itt)  &
 &                     *CONJG(pp%prop(ic,jc,     is,js,isp,itt))
      enddo
      enddo
    enddo
    enddo
  enddo
  this%fV(itt) = REAL(ztmp,kind=DP)
  enddo

#ifndef _singlePU
  call comlib_sumcast(this%fV(:))
#endif
  this%fV(:) = this%fV(:)/REAL(2*NTX*NTY*NTZ,kind=DP)

  call delete(pp)
  deallocate(pp)

  itraj = get_trajectory_number(this%status)

  if (0==nodeid) then
    do itt=2,NTT
      write(str,'(I6,I4,4E24.16)')itraj,itt, &
 &    REAL(this%fV(itt),kind=DP)
      call print(this%logfile_ZV,TRIM(str))
    enddo
  endif

  return
end subroutine

subroutine get_ZA(this,qp,u)
  use comlib, only : comlib_sumcast
  use field_fermion_class, only : igamma, zgamma
  type(sf_measurement),  intent(inout) :: this
  type(quark_prop),      intent(in)    :: qp
  type(vfield_gluon_wg), intent(inout) :: u
  character(len=CHARLEN) :: str
  character(len=1) :: channel_Y,channel_X
  integer :: itraj
  integer :: isp,ic,jc,is,js,itt,jtt
  integer :: ich_Y, ich_X
  integer :: igm5,igm4
  integer :: igmX(SPIN)
  integer :: t_src
  complex(DP) :: zgmX(SPIN)
  complex(DP) :: ztmp0,ztmp1,ztmp2,ztmp3
  real(DP) :: diag_ab(2:NTT,2:NTT,0:1,0:1)
  real(DP) :: diag_cd(2:NTT,2:NTT,0:1,0:1)
  real(DP) ::  diag_e(2:NTT,0:1)
  real(DP) ::  diag_f(2:NTT,0:1)

  type(quark_prop), allocatable :: pp,qc

  if (0 == nodeid) then
    write(*,'("% Start ZA computation")')
    write(*,'("% Nf=",I3)')this%nflavor
  endif

  allocate(pp,qc)
  call new(pp)
  allocate(pp%prop(COL,COL,SPIN,SPIN,NZ*NY*NX,NTT))
  call clear_prop(pp)

  !--------------------------------------------
  ! pp(x,t) = H'(x,t) gamma_5 K     (src_spin=1,2)
  !         = H(x,t) K^\dag gamma_5 (src_spin=3,4)
  !--------------------------------------------
  do itt=2,NTT
!$OMP PARALLEL DO PRIVATE(isp,is,js,ic,jc)
  do isp=1,NSPACE
    !--------------------------------------------
    ! pp(x,t) = H'(x,t) gamma_5 K     (src_spin=1,2)
    !--------------------------------------------
    do js=1,SPIN/2  ! source upper half spin
    do is=1,SPIN
    do jc=1,COL
    do ic=1,COL

      pp%prop(ic,jc,is,js,isp,itt) = ( qp%prop(ic, 1,is, 3,isp,itt)*this%K( 1,jc,igamma(3,5),js)  &
 &                                   + qp%prop(ic, 2,is, 3,isp,itt)*this%K( 2,jc,igamma(3,5),js)  &
 &                                   + qp%prop(ic, 3,is, 3,isp,itt)*this%K( 3,jc,igamma(3,5),js) )*zgamma(3,5) &
 &                                 + ( qp%prop(ic, 1,is, 4,isp,itt)*this%K( 1,jc,igamma(4,5),js)  &
 &                                   + qp%prop(ic, 2,is, 4,isp,itt)*this%K( 2,jc,igamma(4,5),js)  &
 &                                   + qp%prop(ic, 3,is, 4,isp,itt)*this%K( 3,jc,igamma(4,5),js) )*zgamma(4,5)

    enddo
    enddo
    enddo
    enddo

    !--------------------------------------------
    ! pp(x,t) = H(x,t) K^\dag gamma_5 (src_spin=3,4)
    !--------------------------------------------
    do js=SPIN/2+1,SPIN  ! source lower half spin
    do is=1,SPIN
    do jc=1,COL
    do ic=1,COL

      pp%prop(ic,jc,is,js,isp,itt) = ( qp%prop(ic, 1,is, 1,isp,itt)*conjg(this%K(jc, 1,igamma(js,5), 1))  &
 &                                   + qp%prop(ic, 2,is, 1,isp,itt)*conjg(this%K(jc, 2,igamma(js,5), 1))  &
 &                                   + qp%prop(ic, 3,is, 1,isp,itt)*conjg(this%K(jc, 3,igamma(js,5), 1)) )*conjg(zgamma(js,5)) &
 &                                 + ( qp%prop(ic, 1,is, 2,isp,itt)*conjg(this%K(jc, 1,igamma(js,5), 2))  &
 &                                   + qp%prop(ic, 2,is, 2,isp,itt)*conjg(this%K(jc, 2,igamma(js,5), 2))  &
 &                                   + qp%prop(ic, 3,is, 2,isp,itt)*conjg(this%K(jc, 3,igamma(js,5), 2)) )*conjg(zgamma(js,5))

    enddo
    enddo
    enddo
    enddo

  enddo
  enddo



  !------------------------------------------------------------
  ! diag(a,b)
  !
  !  = (2/NV) sum_x Re[tr[ (gamma5 GammaX)  T_Y(x,t;s) K^dag gamma5 H'(x,t)^dag ]]
  !  = (2/NV) sum_x Re[tr[ (gamma5 GammaX)  T_Y(x,t;s) (H'(x,t) gamma5 K)^dag ]]   (spin-1,2)
  !
  ! diag(c,d)
  !
  !  = (2/NV) sum_x Re[tr[ (gamma5 GammaX) T'_Y(x,t;s) gamma5 K H(x,t)^dag ]]
  !  = (2/NV) sum_x Re[tr[ (gamma5 GammaX) T'_Y(x,t;s) (H(x,t) K^dag gamma5)^dag ]]  (spin-3,4)
  !
  ! diag(e)
  !
  !  = (1/NV) sum_x tr[ (gamam5 GammaX) H'(x,t) H'(x,t)^dag] (spin-3,4)  (= 2*f_X'(T-t))
  !
  ! diag(f)
  !
  !  = (1/NV) sum_x tr[ (gamam5 GammaX) H(x,t) H(x,t)^dag] (spin-1,2)    (= 2*f_X(t) )
  !
  !  X,Y = P or A
  !     daig(e) and diag(f) become real number
  !------------------------------------------------------------

  !--------------------------------
  ! connected diagrams (a,b,c,d)
  !--------------------------------
  t_src = NTT/3 + 1
  do jtt = MAX(2,t_src-2),MIN(NTT,t_src+2)

    do ich_Y = 0,1
      select case(ich_Y)
      case(0)
        channel_Y="P"
      case(1)
        channel_Y="A"
      end select

      call new(qc)
      qc%parameter     => qp%parameter
      qc%wavefunc_sous => qp%wavefunc_sous
      qc%smear_sous    = 1
      qc%wavefunc_sink => NULL()

      if (allocated(qc%prop)) deallocate(qc%prop)
      allocate(qc%prop(COL,COL,SPIN,SPIN,NZ*NY*NX,NTT))

      call clear_prop(qc)
      call set_sf_source_connected(qc,qp,jtt,channel_Y)
      call get_prop(qc,u)
      call change_normalization(this,qc)

      do ich_X = 0,1

        select case(ich_X)
        case(0)
          !------------------------------
          ! X = P
          ! GammaX = GammaP = gamma5
          ! gamma5 GammaP = 1
          !------------------------------
          channel_X="P"
          do is=1,SPIN
            igmX(is) = is
            zgmX(is) = Z1
          enddo
        case(1)
          !------------------------------
          ! X = A
          ! GammaX = GammaA = gamma4 gamma5
          ! gamma5 GammaA = - gamma4
          !------------------------------
          channel_X="A"
          do is=1,SPIN
            igmX(is) =  igamma(is,4)
            zgmX(is) = -zgamma(is,4)
          enddo
        end select

        do itt = 2,NTT
          ztmp0 = Z0
          ztmp1 = Z0
!$OMP PARALLEL DO PRIVATE(isp,is,js,ic,jc) REDUCTION(+:ztmp0,ztmp1)
          do isp = 1,NSPACE

            do js=1,SPIN/2
            do is=1,SPIN

              do jc=1,COL
              do ic=1,COL
                ztmp0 = ztmp0 + zgmX(is)*qc%prop(ic,jc,igmX(is),js       ,isp,itt)  &  ! diag(a,b), js=1,2
 &                                *CONJG(pp%prop(ic,jc,      is,js       ,isp,itt))
  
                ztmp1 = ztmp1 + zgmX(is)*qc%prop(ic,jc,igmX(is),js+SPIN/2,isp,itt)  &  ! diag(c,d), js+SPIN/2=3,4
 &                                *CONJG(pp%prop(ic,jc,      is,js+SPIN/2,isp,itt))
              enddo
              enddo

            enddo
            enddo
          enddo

          diag_ab(itt,jtt,ich_X,ich_Y) = REAL(ztmp0,kind=DP)
          diag_cd(itt,jtt,ich_X,ich_Y) = REAL(ztmp1,kind=DP)

        enddo

#ifndef _singlePU
        call comlib_sumcast(diag_ab(:,jtt,ich_X,ich_Y))
        call comlib_sumcast(diag_cd(:,jtt,ich_X,ich_Y))
#endif
        diag_ab(:,jtt,ich_X,ich_Y) = 2*diag_ab(:,jtt,ich_X,ich_Y)/REAL(NTX*NTY*NTZ,kind=DP)
        diag_cd(:,jtt,ich_X,ich_Y) = 2*diag_cd(:,jtt,ich_X,ich_Y)/REAL(NTX*NTY*NTZ,kind=DP)

      enddo ! end of do ich_X
      call delete(qc)
    enddo ! end of do ich_Y

  enddo ! end of do jtt

  !--------------------------------
  ! disconnected diagrams (e,f)
  !--------------------------------
  do ich_X = 0,1

    select case(ich_X)
    case(0)
      !------------------------------
      ! X = P
      ! GammaX = GammaP = gamma5
      ! gamma5 GammaP = 1
      !------------------------------
      channel_X="P"
      do is=1,SPIN
        igmX(is) = is
        zgmX(is) = Z1
      enddo
    case(1)
      !------------------------------
      ! X = A
      ! GammaX = GammaA = gamma4 gamma5
      ! gamma5 GammaA = - gamma4
      !------------------------------
      channel_X="A"
      do is=1,SPIN
        igmX(is) =  igamma(is,4)
        zgmX(is) = -zgamma(is,4)
      enddo
    end select

    do itt = 2,NTT
      ztmp2 = Z0
      ztmp3 = Z0
!$OMP PARALLEL DO PRIVATE(isp,is,js,ic,jc) REDUCTION(+:ztmp2,ztmp3)
      do isp = 1,NSPACE
  
        do js=1,SPIN/2
        do is=1,SPIN

          do jc=1,COL
          do ic=1,COL
            ztmp2 = ztmp2 + zgmX(is)*qp%prop(ic,jc,igmX(is),js+SPIN/2,isp,itt)  &  ! diag(e), js+SPIN/2=3,4
 &                            *CONJG(qp%prop(ic,jc,      is,js+SPIN/2,isp,itt))
 
            ztmp3 = ztmp3 + zgmX(is)*qp%prop(ic,jc,igmX(is),js       ,isp,itt)  &  ! diag(f), js=1,2
 &                            *CONJG(qp%prop(ic,jc,      is,js       ,isp,itt))
          enddo
          enddo

        enddo
        enddo

      enddo

      diag_e(itt,ich_X) = REAL(ztmp2,kind=DP)
      diag_f(itt,ich_X) = REAL(ztmp3,kind=DP)

    enddo  ! end of do itt

#ifndef _singlePU
    call comlib_sumcast(diag_e(:,ich_X))
    call comlib_sumcast(diag_f(:,ich_X))
#endif
    diag_e(:,ich_X) = diag_e(:,ich_X)/REAL(NTX*NTY*NTZ,kind=DP)
    diag_f(:,ich_X) = diag_f(:,ich_X)/REAL(NTX*NTY*NTZ,kind=DP)

  enddo ! end of do ich_X

  call delete(qc)
  call delete(pp)
  deallocate(pp,qc)

  !----------------------------------------------------------
  ! Store the results
  !----------------------------------------------------------
  itraj = get_trajectory_number(this%status)

  !----------------------------------------------------------------
  ! correlation functions
  !  fAA,fAP,fPA,fPP : full contranction (connected+disconnected)
  ! 
  !  fAAc,fAPc,fPAc,fPPc : partial contranction (connected only)
  !----------------------------------------------------------------
  this%fAA(:,:)  = Z0
  this%fAP(:,:)  = Z0
  this%fPA(:,:)  = Z0
  this%fPP(:,:)  = Z0
  this%fAAc(:,:) = Z0
  this%fAPc(:,:) = Z0
  this%fPAc(:,:) = Z0
  this%fPPc(:,:) = Z0

  do jtt = MAX(2,t_src-2),MIN(NTT,t_src+2)
  do itt = 2,NTT
    this%fAA(itt,jtt) = (  diag_ab(itt,jtt,1,1)                               &
 &                        -diag_cd(itt,jtt,1,1)                               &
 &                       -2*diag_e(itt,1)*diag_f(jtt,1)/this%nflavor  &
 &                       +2*diag_f(itt,1)*diag_e(jtt,1)/this%nflavor )/4.0_DP
    this%fAP(itt,jtt) = (  diag_ab(itt,jtt,1,0)                               &
 &                        -diag_cd(itt,jtt,1,0)                               &
 &                       -2*diag_e(itt,1)*diag_f(jtt,0)/this%nflavor  &
 &                       +2*diag_f(itt,1)*diag_e(jtt,0)/this%nflavor )/4.0_DP
    this%fPA(itt,jtt) = (  diag_ab(itt,jtt,0,1)                               &
 &                        -diag_cd(itt,jtt,0,1)                               &
 &                       -2*diag_e(itt,0)*diag_f(jtt,1)/this%nflavor  &
 &                       +2*diag_f(itt,0)*diag_e(jtt,1)/this%nflavor )/4.0_DP
    this%fPP(itt,jtt) = (  diag_ab(itt,jtt,0,0)                               &
 &                        -diag_cd(itt,jtt,0,0)                               &
 &                       -2*diag_e(itt,0)*diag_f(jtt,0)/this%nflavor  &
 &                       +2*diag_f(itt,0)*diag_e(jtt,0)/this%nflavor )/4.0_DP

    this%fAAc(itt,jtt) = ( diag_ab(itt,jtt,1,1) - diag_cd(itt,jtt,1,1) )/4.0_DP
    this%fAPc(itt,jtt) = ( diag_ab(itt,jtt,1,0) - diag_cd(itt,jtt,1,0) )/4.0_DP
    this%fPAc(itt,jtt) = ( diag_ab(itt,jtt,0,1) - diag_cd(itt,jtt,0,1) )/4.0_DP
    this%fPPc(itt,jtt) = ( diag_ab(itt,jtt,0,0) - diag_cd(itt,jtt,0,0) )/4.0_DP
  enddo
  enddo

  if (0==nodeid) then
    write(str,'("# full contraction (connected + disconnected), Nf=",I3)')this%nflavor
    call print(this%logfile_ZA,TRIM(str))
    write(str,'("# traj itt jtt",7X,"fAA(itt,jtt)",12X,"fAP(itt,jtt)",12X,"fPA(itt,jtt)",12X,"fPP(itt,jtt)")')
    call print(this%logfile_ZA,TRIM(str))
    do jtt = MAX(2,t_src-2),MIN(NTT,t_src+2)
    do itt = 2,NTT
      write(str,'(I6,2I4,8E24.16)')itraj,itt,jtt,  &
 &                REAL(this%fAA(itt,jtt),kind=DP), &
 &                REAL(this%fAP(itt,jtt),kind=DP), &
 &                REAL(this%fPA(itt,jtt),kind=DP), &
 &                REAL(this%fPP(itt,jtt),kind=DP)
      call print(this%logfile_ZA,TRIM(str))
    enddo
    enddo

    write(str,'("# partial contraction (connected only)")')
    call print(this%logfile_ZA,TRIM(str))
    write(str,'("# traj itt jtt",7X,"fAA(itt,jtt)",12X,"fAP(itt,jtt)",12X,"fPA(itt,jtt)",12X,"fPP(itt,jtt)")')
    call print(this%logfile_ZA,TRIM(str))
    do jtt = MAX(2,t_src-2),MIN(NTT,t_src+2)
    do itt = 2,NTT
      write(str,'(I6,2I4,8E24.16)')itraj,itt,jtt,   &
 &                REAL(this%fAAc(itt,jtt),kind=DP), &
 &                REAL(this%fAPc(itt,jtt),kind=DP), &
 &                REAL(this%fPAc(itt,jtt),kind=DP), &
 &                REAL(this%fPPc(itt,jtt),kind=DP)
      call print(this%logfile_ZA,TRIM(str))
    enddo
    enddo
  endif


  do ich_X = 0,1

    select case(ich_X)
    case(0)
      channel_X="P"
    case(1)
      channel_X="A"
    end select

    do ich_Y = 0,1
  
      select case(ich_Y)
      case(0)
        channel_Y="P"
      case(1)
        channel_Y="A"
      end select

      if (0==nodeid) then

        write(str,'("# Gamma_X=Gamma_",A," Gamma_Y=Gamma_",A)')channel_X,channel_Y
        call print(this%logfile_ZA,TRIM(str))
        write(str,'("# traj itt jtt",4X,"diag(a,b)(itt,jtt)",6X,"diag(c,d)(itt,jtt)")')
        call print(this%logfile_ZA,TRIM(str))

        do jtt = MAX(2,t_src-2),MIN(NTT,t_src+2)
        do itt = 2,NTT
          write(str,'(I6,2I4,8E24.16)')itraj,itt,jtt,diag_ab(itt,jtt,ich_X,ich_Y),diag_cd(itt,jtt,ich_X,ich_Y)
          call print(this%logfile_ZA,TRIM(str))
        enddo
        enddo

      endif

    enddo ! end of do ich_Y

    if (0==nodeid) then

      write(str,'("# Gamma_X=Gamma_",A)')channel_X
      call print(this%logfile_ZA,TRIM(str))
      write(str,'("# traj itt",8X,"diag(e)(itt)",12X,"diag(f)(itt)")')
      call print(this%logfile_ZA,TRIM(str))

      do itt=2,NTT
        write(str,'(I6,I4,8E24.16)')itraj,itt,diag_e(itt,ich_X),diag_f(itt,ich_X)
        call print(this%logfile_ZA,TRIM(str))
      enddo

    endif

  enddo ! end of do ich_X

  return
end subroutine


function get_gsf_normalization(gauge) result(knorm)
!
! Normalizetion from gauge action on a classical SF field
!
  use hmc_gluon_class, only : gluon_rg_sf
  implicit none
  type(gluon_rg_sf), intent(in) :: gauge
  real(DP) :: knorm
  real(DP) :: th(NDIM-1)

  th(1) = PI/COL/NTT/NTX
  th(2) = PI/COL/NTT/NTY
  th(3) = PI/COL/NTT/NTZ
  
#ifdef _RGGAUGE
  knorm=REAL(4*NTX*NTY*NTZ,kind=DP)*( gauge%cRG(0)*( (sin(2*th(1))+sin(th(1)))/NTX     &
 &                                                  +(sin(2*th(2))+sin(th(2)))/NTY     &
 &                                                  +(sin(2*th(3))+sin(th(3)))/NTZ)    &
 &                                 +4*gauge%cRG(1)*( (sin(4*th(1))+sin(2*th(1)))/NTX   &
 &                                                  +(sin(4*th(2))+sin(2*th(2)))/NTY   &
 &                                                  +(sin(4*th(3))+sin(2*th(3)))/NTZ) )
#else
  knorm=REAL(4*NTX*NTY*NTZ,kind=DP)*( (sin(2*th(1))+sin(th(1)))/NTX    &
 &                                   +(sin(2*th(2))+sin(th(2)))/NTY    &
 &                                   +(sin(2*th(3))+sin(th(3)))/NTZ )
#endif

 return
end function


subroutine get_sf_scheme_coupling(this,gauge,u)
!
! Calculate SF scheme coupling
!
!  d Gamma|                  1     
!  -------|         = k ( -------- - mu v )
!  d  eta |_eta=0         (g_SF)^2 
!
!     k : normalization contant depending on the gauge action
!
!  this : SF parameters
! gauge : gauge action parameters
!     u : guage link field
!
!
  use comlib, only : comlib_sumcast
  use hmc_gluon_class, only : gluon_rg_sf
  implicit none
  type(sf_measurement), intent(inout) :: this
  type(gluon_rg_sf),       intent(in) :: gauge
  type(vfield_gluon_wg),   intent(in) :: u

  integer :: itraj
  character(len=CHARLEN) :: str

  real(DP)    :: dSGg,dSGv
  complex(DP) :: w23(COL,COL),w41(COL,COL)
  complex(DP) :: w25(COL,COL),w61(COL,COL)
  complex(DP) :: w2341(COL,COL),w2561(COL,COL)
  complex(DP) :: uk0(COL,COL),vk0(COL,COL)
  complex(DP), parameter :: zi=(0.0_DP,1.0_DP)
  real(DP) :: omega(COL,NDIM-1),vi(COL,NDIM-1)
  real(DP) :: knorm,th(NDIM-1)
  real(DP) :: trwUk0,trwVk0
  integer :: ieo,ioe,ix,iy,iz,ieoxyz,ic,jc,mu
  integer :: ix2,iy2,iz2
#ifdef _RGGAUGE
  real(DP) :: dSGg1,dSGv1,dSGg2,dSGv2
  complex(DP) :: we4(COL,COL),we41(COL,COL),w23d(COL,COL)
  complex(DP) :: wi6(COL,COL),wi61(COL,COL),w25h(COL,COL)
  complex(DP) :: w23de41(COL,COL),w25hi61(COL,COL)
  complex(DP) :: w78(COL,COL),w478(COL,COL),w123(COL,COL)
  complex(DP) :: w98(COL,COL),w698(COL,COL),w125(COL,COL)
  complex(DP) :: w123478(COL,COL),w125698(COL,COL)
  integer :: ix8,iy8,iz8
#endif

  omega(1,1)= 1.0_DP/NTX
  omega(1,2)= 1.0_DP/NTY
  omega(1,3)= 1.0_DP/NTZ
  omega(2,1)=-0.5_DP/NTX
  omega(2,2)=-0.5_DP/NTY
  omega(2,3)=-0.5_DP/NTZ
  omega(3,1)=-0.5_DP/NTX
  omega(3,2)=-0.5_DP/NTY
  omega(3,3)=-0.5_DP/NTZ

  vi(1,1)= 0.0_DP
  vi(1,2)= 0.0_DP
  vi(1,3)= 0.0_DP
  vi(2,1)= 1.0_DP/NTX
  vi(2,2)= 1.0_DP/NTY
  vi(2,3)= 1.0_DP/NTZ
  vi(3,1)=-1.0_DP/NTX
  vi(3,2)=-1.0_DP/NTY
  vi(3,3)=-1.0_DP/NTZ

  !
  ! gauge action normalizetion knorm
  !   from action on a classical SF field
  !
  knorm = get_gsf_normalization(gauge) 

#ifdef _GGGG_

  th(1) = PI/COL/NTT/NTX
  th(2) = PI/COL/NTT/NTY
  th(3) = PI/COL/NTT/NTZ

#ifdef _RGGAUGE
  knorm=REAL(4*NTX*NTY*NTZ,kind=DP)*( gauge%cRG(0)*( (sin(2*th(1))+sin(th(1)))/NTX     &
 &                                                  +(sin(2*th(2))+sin(th(2)))/NTY     &
 &                                                  +(sin(2*th(3))+sin(th(3)))/NTZ)    &
 &                                 +4*gauge%cRG(1)*( (sin(4*th(1))+sin(2*th(1)))/NTX   &
 &                                                  +(sin(4*th(2))+sin(2*th(2)))/NTY   &
 &                                                  +(sin(4*th(3))+sin(2*th(3)))/NTZ) )
#else
  knorm=REAL(4*NTX*NTY*NTZ,kind=DP)*( (sin(2*th(1))+sin(th(1)))/NTX    &
 &                                   +(sin(2*th(2))+sin(th(2)))/NTY    &
 &                                   +(sin(2*th(3))+sin(th(3)))/NTZ )
#endif
#endif

  dSGg=0.0_DP  ! for 1/gSF^2
  dSGv=0.0_DP  ! for v
#ifdef _RGGAUGE
  dSGg1=0.0_DP  ! for 1/gSF^2 (rectangular part 1x2)
  dSGv1=0.0_DP  ! for v       (rectangular part 1x2)
  dSGg2=0.0_DP  ! for 1/gSF^2 (rectangular part 2x1)
  dSGv2=0.0_DP  ! for v       (rectangular part 2x1)
#endif

#ifdef _RGGAUGE
!$OMP PARALLEL DO &
!$OMP& PRIVATE(ix,iy,iz,ieoxyz,ieo,ioe,mu,ic,jc, &
!$OMP&         ix2,iy2,iz2,w23,w25,w41,w61,w2341,w2561,uk0,vk0,trwUk0,trwVk0, &
!$OMP&         ix8,iy8,iz8,we4,wi6,we41,w23d,wi61,w25h,w23de41,w25hi61, &
!$OMP&                     w78,w98,w478,w123,w698,w125,w123478,w125698) &
!$OMP& REDUCTION(+:dSGg,dSGv,dSGg1,dSGv1,dSGg2,dSGv2)
#else
!$OMP PARALLEL DO &
!$OMP& PRIVATE(ix,iy,iz,ieoxyz,ieo,ioe,mu,ic,jc, &
!$OMP&         ix2,iy2,iz2,w23,w25,w41,w61,w2341,w2561,uk0,vk0,trwUk0,trwVk0) &
!$OMP& REDUCTION(+:dSGg,dSGv)
#endif
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz,2)

    do mu=1,NDIM-1  !<=== Spatial only
      ix2=ix+isx(mu) ! n + e_k
      iy2=iy+isy(mu)
      iz2=iz+isz(mu)
#ifdef _RGGAUGE
      ix8=ix-isx(mu) ! n - e_k
      iy8=iy-isy(mu)
      iz8=iz-isz(mu)
#endif
      if (1 == ieoxyz) then
        ieo = 0
        ioe = 1-ieo
      else
        ieo = 1
        ioe = 1-ieo
      endif

!******************************************
!*     v                  t=NTT+1     v
!*   n+4                  n         n+k
!*    4---<---3            1--->---2
!*    |       |            |       |
!*    v       ^            ^       v
!*    |       |            |       |
!*    1--->---2  v         6---<---5
!*   n         n+k       n-4
!*   t = 1                          for plaquette
!******************************************
      do jc=1,COL
      do ic=1,COL
        w23(ic,jc)=       u%eo(ieo)%mu(mu)%s(  0,iz ,iy ,ix )%u(ic, 1)*   &
 &                        u%eo(ioe)%mu( 4)%s(  0,iz2,iy2,ix2)%u( 1,jc)    &
 &                       +u%eo(ieo)%mu(mu)%s(  0,iz ,iy ,ix )%u(ic, 2)*   &
 &                        u%eo(ioe)%mu( 4)%s(  0,iz2,iy2,ix2)%u( 2,jc)    &
 &                       +u%eo(ieo)%mu(mu)%s(  0,iz ,iy ,ix )%u(ic, 3)*   &
 &                        u%eo(ioe)%mu( 4)%s(  0,iz2,iy2,ix2)%u( 3,jc)
        w25(ic,jc)=       u%eo(ieo)%mu(mu)%s(NTH,iz ,iy ,ix )%u(ic, 1)*   &
 &                  conjg(u%eo(ieo)%mu( 4)%s(NTH,iz2,iy2,ix2)%u(jc, 1))   &
 &                 +      u%eo(ieo)%mu(mu)%s(NTH,iz ,iy ,ix )%u(ic, 2)*   &
 &                  conjg(u%eo(ieo)%mu( 4)%s(NTH,iz2,iy2,ix2)%u(jc, 2))   &
 &                 +      u%eo(ieo)%mu(mu)%s(NTH,iz ,iy ,ix )%u(ic, 3)*   &
 &                  conjg(u%eo(ieo)%mu( 4)%s(NTH,iz2,iy2,ix2)%u(jc, 3))
      enddo
      enddo

      do jc=1,COL
      do ic=1,COL
        w41(ic,jc)= conjg(u%eo(ioe)%mu(mu)%s(  1,iz,iy,ix)%u( 1,ic))*   &
 &                  conjg(u%eo(ieo)%mu( 4)%s(  0,iz,iy,ix)%u(jc, 1))    &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(  1,iz,iy,ix)%u( 2,ic))*   &
 &                  conjg(u%eo(ieo)%mu( 4)%s(  0,iz,iy,ix)%u(jc, 2))    &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(  1,iz,iy,ix)%u( 3,ic))*   &
 &                  conjg(u%eo(ieo)%mu( 4)%s(  0,iz,iy,ix)%u(jc, 3))
        w61(ic,jc)= conjg(u%eo(ioe)%mu(mu)%s(NTH,iz,iy,ix)%u( 1,ic))*   &
 &                        u%eo(ioe)%mu( 4)%s(NTH,iz,iy,ix)%u( 1,jc)     &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(NTH,iz,iy,ix)%u( 2,ic))*   &
 &                        u%eo(ioe)%mu( 4)%s(NTH,iz,iy,ix)%u( 2,jc)     &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(NTH,iz,iy,ix)%u( 3,ic))*   &
 &                        u%eo(ioe)%mu( 4)%s(NTH,iz,iy,ix)%u( 3,jc)
      enddo
      enddo

      do jc=1,COL
      do ic=1,COL
        w2341(ic,jc)=w23(ic,1)*w41(1,jc)  &
 &                  +w23(ic,2)*w41(2,jc)  &
 &                  +w23(ic,3)*w41(3,jc)
        w2561(ic,jc)=w25(ic,1)*w61(1,jc)  &
 &                  +w25(ic,2)*w61(2,jc)  &
 &                  +w25(ic,3)*w61(3,jc)
      enddo
      enddo
      do jc=1,COL
      do ic=1,COL
        uk0(ic,jc)=(w2341(ic,jc)-conjg(w2341(jc,ic)))  &
 &                -(w2561(ic,jc)-conjg(w2561(jc,ic)))
        vk0(ic,jc)=(w2341(ic,jc)-conjg(w2341(jc,ic)))  &
 &                +(w2561(ic,jc)-conjg(w2561(jc,ic)))
      enddo
      enddo
      trwUk0 = -REAL(ZI*( omega(1,mu)*uk0(1,1)+omega(2,mu)*uk0(2,2)+omega(3,mu)*uk0(3,3)),kind=DP)
      trwVk0 = -REAL(ZI*(    vi(1,mu)*vk0(1,1)+   vi(2,mu)*vk0(2,2)+   vi(3,mu)*vk0(3,3)),kind=DP)
      dSGg=dSGg+trwUk0
      dSGv=dSGv+trwVk0


#ifdef _RGGAUGE
!******************************************
!*
!*     v                  t=NTT+1     v
!* n+2*4                  n         n+k
!*    e---<---d            1--->---2
!*    |       |            |       |
!*    v       ^            ^       v
!*    |       |            |       |
!*    4       3            6       5
!*    |       |            |       |
!*    v       ^            ^       v
!*    |       |            |       |
!*    1--->---2  v         i---<---h
!*   n         n+k       n-2*4
!*   t = 1                          for rectangle (RG)
!*               v           v        t=NTT+1    v
!*             n+4         n-k        n        n+k
!*     7---<---4---<---3      8--->---1--->---2
!*     |               |      |               |
!*     v               ^      ^               v
!*     |               |      |               |
!*   v 8--->---1--->---2  v   9---<---6---<---5
!* n-k         n        n+k          n-4
!*             t = 1                for rectangle (RG)
!*
!******************************************
      do jc=1,COL
      do ic=1,COL
        we4(ic,jc)= conjg(u%eo(ieo)%mu(mu)%s(    1,iz,iy,ix)%u( 1,ic))*   &
 &                  conjg(u%eo(ioe)%mu( 4)%s(    1,iz,iy,ix)%u(jc, 1))    &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(    1,iz,iy,ix)%u( 2,ic))*   &
 &                  conjg(u%eo(ioe)%mu( 4)%s(    1,iz,iy,ix)%u(jc, 2))    &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(    1,iz,iy,ix)%u( 3,ic))*   &
 &                  conjg(u%eo(ioe)%mu( 4)%s(    1,iz,iy,ix)%u(jc, 3))
        wi6(ic,jc)= conjg(u%eo(ieo)%mu(mu)%s(NTH-1,iz,iy,ix)%u( 1,ic))*   &
 &                        u%eo(ieo)%mu( 4)%s(NTH-1,iz,iy,ix)%u( 1,jc)     &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(NTH-1,iz,iy,ix)%u( 2,ic))*   &
 &                        u%eo(ieo)%mu( 4)%s(NTH-1,iz,iy,ix)%u( 2,jc)     &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(NTH-1,iz,iy,ix)%u( 3,ic))*   &
 &                        u%eo(ieo)%mu( 4)%s(NTH-1,iz,iy,ix)%u( 3,jc)
      enddo
      enddo

      do jc=1,COL
      do ic=1,COL
        we41(ic,jc)=we4(ic,1)*conjg(u%eo(ieo)%mu( 4)%s(    0,iz ,iy ,ix )%u(jc, 1))  &
 &                 +we4(ic,2)*conjg(u%eo(ieo)%mu( 4)%s(    0,iz ,iy ,ix )%u(jc, 2))  &
 &                 +we4(ic,3)*conjg(u%eo(ieo)%mu( 4)%s(    0,iz ,iy ,ix )%u(jc, 3))
        w23d(ic,jc)=w23(ic,1)*      u%eo(ieo)%mu( 4)%s(    1,iz2,iy2,ix2)%u( 1,jc)   &
 &                 +w23(ic,2)*      u%eo(ieo)%mu( 4)%s(    1,iz2,iy2,ix2)%u( 2,jc)   &
 &                 +w23(ic,3)*      u%eo(ieo)%mu( 4)%s(    1,iz2,iy2,ix2)%u( 3,jc)
        wi61(ic,jc)=wi6(ic,1)*      u%eo(ioe)%mu( 4)%s(NTH  ,iz ,iy ,ix )%u( 1,jc)   &
 &                 +wi6(ic,2)*      u%eo(ioe)%mu( 4)%s(NTH  ,iz ,iy ,ix )%u( 2,jc)   &
 &                 +wi6(ic,3)*      u%eo(ioe)%mu( 4)%s(NTH  ,iz ,iy ,ix )%u( 3,jc)
        w25h(ic,jc)=w25(ic,1)*conjg(u%eo(ioe)%mu( 4)%s(NTH-1,iz2,iy2,ix2)%u(jc, 1))  &
 &                 +w25(ic,2)*conjg(u%eo(ioe)%mu( 4)%s(NTH-1,iz2,iy2,ix2)%u(jc, 2))  &
 &                 +w25(ic,3)*conjg(u%eo(ioe)%mu( 4)%s(NTH-1,iz2,iy2,ix2)%u(jc, 3))
      enddo
      enddo

      do jc=1,COL
      do ic=1,COL
        w23de41(ic,jc)=w23d(ic,1)*we41(1,jc)  &
 &                    +w23d(ic,2)*we41(2,jc)  &
 &                    +w23d(ic,3)*we41(3,jc)
        w25hi61(ic,jc)=w25h(ic,1)*wi61(1,jc)  &
 &                    +w25h(ic,2)*wi61(2,jc)  &
 &                    +w25h(ic,3)*wi61(3,jc)
      enddo
      enddo
      do jc=1,COL
      do ic=1,COL
        uk0(ic,jc)=(w23de41(ic,jc)-conjg(w23de41(jc,ic)))  &
 &                -(w25hi61(ic,jc)-conjg(w25hi61(jc,ic)))
        vk0(ic,jc)=(w23de41(ic,jc)-conjg(w23de41(jc,ic)))  &
 &                +(w25hi61(ic,jc)-conjg(w25hi61(jc,ic)))
      enddo
      enddo
      trwUk0 = -REAL(ZI*( omega(1,mu)*uk0(1,1)+omega(2,mu)*uk0(2,2)+omega(3,mu)*uk0(3,3)),kind=DP)
      trwVk0 = -REAL(ZI*(    vi(1,mu)*vk0(1,1)+   vi(2,mu)*vk0(2,2)+   vi(3,mu)*vk0(3,3)),kind=DP)
      dSGg1=dSGg1+trwUk0
      dSGv1=dSGv1+trwVk0


      do jc=1,COL
      do ic=1,COL
        w78(ic,jc)= conjg(u%eo(ieo)%mu(mu)%s(  1,iz8,iy8,ix8)%u( 1,ic))*   &
 &                  conjg(u%eo(ioe)%mu( 4)%s(  0,iz8,iy8,ix8)%u(jc, 1))    &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(  1,iz8,iy8,ix8)%u( 2,ic))*   &
 &                  conjg(u%eo(ioe)%mu( 4)%s(  0,iz8,iy8,ix8)%u(jc, 2))    &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(  1,iz8,iy8,ix8)%u( 3,ic))*   &
 &                  conjg(u%eo(ioe)%mu( 4)%s(  0,iz8,iy8,ix8)%u(jc, 3))
        w98(ic,jc)= conjg(u%eo(ieo)%mu(mu)%s(NTH,iz8,iy8,ix8)%u( 1,ic))*   &
 &                        u%eo(ieo)%mu( 4)%s(NTH,iz8,iy8,ix8)%u( 1,jc)     &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(NTH,iz8,iy8,ix8)%u( 2,ic))*   &
 &                        u%eo(ieo)%mu( 4)%s(NTH,iz8,iy8,ix8)%u( 2,jc)     &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(NTH,iz8,iy8,ix8)%u( 3,ic))*   &
 &                        u%eo(ieo)%mu( 4)%s(NTH,iz8,iy8,ix8)%u( 3,jc)
      enddo
      enddo

      do jc=1,COL
      do ic=1,COL
        w478(ic,jc)=conjg(u%eo(ioe)%mu(mu)%s(  1,iz ,iy ,ix )%u( 1,ic))*w78(1,jc)   &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(  1,iz ,iy ,ix )%u( 2,ic))*w78(2,jc)   &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(  1,iz ,iy ,ix )%u( 3,ic))*w78(3,jc)
        w123(ic,jc)=      u%eo(ioe)%mu(mu)%s(  0,iz8,iy8,ix8)%u(ic, 1) *w23(1,jc)   &
 &                 +      u%eo(ioe)%mu(mu)%s(  0,iz8,iy8,ix8)%u(ic, 2) *w23(2,jc)   &
 &                 +      u%eo(ioe)%mu(mu)%s(  0,iz8,iy8,ix8)%u(ic, 3) *w23(3,jc)
        w698(ic,jc)=conjg(u%eo(ioe)%mu(mu)%s(NTH,iz ,iy ,ix )%u( 1,ic))*w98(1,jc)   &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(NTH,iz ,iy ,ix )%u( 2,ic))*w98(2,jc)   &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(NTH,iz ,iy ,ix )%u( 3,ic))*w98(3,jc)
        w125(ic,jc)=      u%eo(ioe)%mu(mu)%s(NTH,iz8,iy8,ix8)%u(ic, 1) *w25(1,jc)   &
 &                 +      u%eo(ioe)%mu(mu)%s(NTH,iz8,iy8,ix8)%u(ic, 2) *w25(2,jc)   &
 &                 +      u%eo(ioe)%mu(mu)%s(NTH,iz8,iy8,ix8)%u(ic, 3) *w25(3,jc)
      enddo
      enddo

      do jc=1,COL
      do ic=1,COL
        w123478(ic,jc)=w123(ic,1)*w478(1,jc)  &
 &                    +w123(ic,2)*w478(2,jc)  &
 &                    +w123(ic,3)*w478(3,jc)
        w125698(ic,jc)=w125(ic,1)*w698(1,jc)  &
 &                    +w125(ic,2)*w698(2,jc)  &
 &                    +w125(ic,3)*w698(3,jc)
      enddo
      enddo
      do jc=1,COL
      do ic=1,COL
        uk0(ic,jc)=(w123478(ic,jc)-conjg(w123478(jc,ic)))  &
 &                -(w125698(ic,jc)-conjg(w125698(jc,ic)))
        vk0(ic,jc)=(w123478(ic,jc)-conjg(w123478(jc,ic)))  &
 &                +(w125698(ic,jc)-conjg(w125698(jc,ic)))
      enddo
      enddo
      trwUk0 = -REAL(ZI*( omega(1,mu)*uk0(1,1)+omega(2,mu)*uk0(2,2)+omega(3,mu)*uk0(3,3)),kind=DP)
      trwVk0 = -REAL(ZI*(    vi(1,mu)*vk0(1,1)+   vi(2,mu)*vk0(2,2)+   vi(3,mu)*vk0(3,3)),kind=DP)
      dSGg2=dSGg2+trwUk0
      dSGv2=dSGv2+trwVk0
#endif

    enddo ! end of do mu
  enddo
  enddo
  enddo  ! end of do ix,iy,iz

#ifndef _singlePU
  call comlib_sumcast(dSGg)
  call comlib_sumcast(dSGv)
#ifdef _RGGAUGE
  call comlib_sumcast(dSGg1)
  call comlib_sumcast(dSGv1)
  call comlib_sumcast(dSGg2)
  call comlib_sumcast(dSGv2)
#endif
#endif

#ifdef _RGGAUGE
  dSGg=(gauge%beta/6.0_DP)*(gauge%ct *gauge%cRG(0)*dSGg   &
 &                         +          gauge%cRG(1)*dSGg1  &
 &                     +2*gauge%ct_rg*gauge%cRG(1)*dSGg2)
  dSGv=(gauge%beta/6.0_DP)*(gauge%ct *gauge%cRG(0)*dSGv   &
 &                         +          gauge%cRG(1)*dSGv1  &
 &                     +2*gauge%ct_rg*gauge%cRG(1)*dSGv2)
#else
  dSGg=(gauge%beta/6.0_DP)*gauge%ct*dSGg
  dSGv=(gauge%beta/6.0_DP)*gauge%ct*dSGv
#endif

  this%gSF2inv =  dSGg/knorm
  this%vSF     = -dSGv/knorm

  itraj = get_trajectory_number(this%status)
  if (0 == nodeid) then
    write(str,'(I6,9E24.15)')itraj,this%gSF2inv,this%vSF,dSGg,-dSGv,knorm
    call print(this%logfile_gsf,TRIM(str))
  endif

  return
end subroutine

subroutine set_z2_noise_fq_sf(y)
!
! set spin-color diluted complex Z2 noise on wavefunction
!
  implicit none
  type(field_quark_wg), intent(inout) :: y
  call set_z2_noise(y%eo(0))
  call set_z2_noise(y%eo(1))
  return
end subroutine

subroutine set_z2_noise_fq_sf_eo(ye)
!
! set spin-color diluted complex Z2 noise on wavefunction
!
  implicit none
  type(field_quark_eo_wg), intent(inout) :: ye
  integer :: iz,iy,ix,itb,ic,is,ieo,ieoxyz
  complex(DP) :: ctmp
  real(DP) :: vec(2),cr,ci
#ifdef _PARALLELCHECK
  integer :: itx,ity,itz,itt,ipx,ipy,ipz
#endif

  !
  ! set spin-color diluted Z2 noise vector
  !
#ifdef _PARALLELCHECK
  !***********************************
  !* Programmed independent on
  !* parallelization size.
  !***********************************
  do itx=1,NTX
  do ity=1,NTY
  do itz=1,NTZ
  do itt=1,NTT
    ieo=mod(itx+ity+itz+itt,2)
    ipx = (itx-1)/NX
    ipy = (ity-1)/NY
    ipz = (itz-1)/NZ
    ix  = mod(itx-1,NX)+1
    iy  = mod(ity-1,NY)+1
    iz  = mod(itz-1,NZ)+1
    itb = itt/2

    call get_rand3(m_rand,vec)

    if ( (ipx == ipsite(1)) .and. &
 &       (ipy == ipsite(2)) .and. &
 &       (ipz == ipsite(3)) .and. &
 &       (ieo == ye%ieo   ) ) then

      if (vec(1) < 0.5_DP) then
        cr=+1.0_DP
      else
        cr=-1.0_DP
      endif
      if (vec(2) < 0.5_DP) then
        ci=+1.0_DP
      else
        ci=-1.0_DP
      endif

      ctmp=cmplx(cr,ci,kind=KIND(ctmp))/SQRT(2.0_DP)

      do is=1,SPIN
      do ic=1,COL
        ye%s(itb,iz,iy,ix)%y(ic,is) = ctmp
      enddo
      enddo
 
    endif
  enddo
  enddo
  enddo
  enddo
#else
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ye%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz

      call get_rand3(m_rand,vec)

      if (vec(1) < 0.5_DP) then
        cr=+1.0_DP
      else
        cr=-1.0_DP
      endif
      if (vec(2) < 0.5_DP) then
        ci=+1.0_DP
      else
        ci=-1.0_DP
      endif

      ctmp=cmplx(cr,ci,kind=KIND(ctmp))/SQRT(2.0_DP)

      do is=1,SPIN
      do ic=1,COL
        ye%s(itb,iz,iy,ix)%y(ic,is) = ctmp
      enddo
      enddo

    enddo
  enddo
  enddo
  enddo ! end of do ixyz
#endif

  call copy_boundary(ye)

#ifdef _SF
  call set_sf_boundary(ye)
#endif

  return
end subroutine

subroutine set_z2_noise_source(y,z,jc,js)
!
! set spin-color diluted complex Z2 noise on full field
!
  type(field_quark_wg), intent(inout) :: y
  type(field_quark_wg), intent(in)    :: z
  integer, intent(in)                 :: jc,js
  call set_z2_noise_source_eo(y%eo(0),z%eo(0),jc,js)
  call set_z2_noise_source_eo(y%eo(1),z%eo(1),jc,js)
  return
end subroutine

subroutine set_z2_noise_source_eo(ye,ze,jc,js)
!
! set spin-color diluted complex Z2 noise on even/odd field
!
  use error_class
  implicit none
  type(field_quark_eo_wg), intent(inout) :: ye
  type(field_quark_eo_wg), intent(in)    :: ze
  integer,                 intent(in)    :: jc,js
  integer :: ix,iy,iz,ieoxyz,itb,ic,is
  character(len=CHARLEN) :: str

  if ( .not. (ye%ieo == ze%ieo) ) then
    write(str,'("even/odd error stop at set_z2_noise_source_eo:",A,I5)')__FILE__,__LINE__
    call error_stop(str)
  endif

  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ye%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      do is=1,SPIN
      do ic=1,COL
        ye%s(itb,iz,iy,ix)%y(ic,is) = Z0
      enddo
      enddo
      ye%s(itb,iz,iy,ix)%y(jc,js) = ze%s(itb,iz,iy,ix)%y(jc,js)
    enddo
  enddo
  enddo
  enddo

  call copy_boundary(ye)

#ifdef _SF
  call set_sf_boundary(ye)
#endif

  return
end subroutine

subroutine set_coordinate_basis_fq_sf(y,itx,ity,itz,itt)
!
! set spin-color diluted coordinate basis vector
!
! y has 1 only at (itx,ity,itz,itt)
!
  implicit none
  type(field_quark_wg), intent(inout) :: y
  integer, intent(in) :: itx,ity,itz,itt
  call set_coordinate_basis(y%eo(0),itx,ity,itz,itt)
  call set_coordinate_basis(y%eo(1),itx,ity,itz,itt)
  return
end subroutine

subroutine set_coordinate_basis_fq_sf_eo(ye,itx,ity,itz,itt)
!
! set spin-color diluted coordinate basis vector
!
! ye has 1 only at (itx,ity,itz,itt)
!
  use error_class
  implicit none
  type(field_quark_eo_wg), intent(inout) :: ye
  integer, intent(in) :: itx,ity,itz,itt

  integer :: iz,iy,ix,itb,ic,is,ieo,ieoxyz
  integer :: ipx,ipy,ipz
  character(len=CHARLEN) :: str

  if ( itx <= 0 .OR. NTX < itt .OR. &
 &     ity <= 0 .OR. NTY < itt .OR. &
 &     itz <= 0 .OR. NTZ < itt .OR. &
 &     itt <= 0 .OR. NTT < itt )  then
    write(str,'(A," at ",I3," : (itx,ity,itz,itt) range error. (",I3,",",I3,",",I3,",",I3,"). STOP.")') &
 &  __FILE__,__LINE__,itx,ity,itz,itt
    call error_stop(str)
  endif

  call clear(ye)

  ieo = mod(itx+ity+itz+itt,2)
  ipx = (itx-1)/NX
  ipy = (ity-1)/NY
  ipz = (itz-1)/NZ
  ix  = mod(itx-1,NX)+1
  iy  = mod(ity-1,NY)+1
  iz  = mod(itz-1,NZ)+1
  itb = itt/2

  if ( (ipx == ipsite(1)) .and. &
 &     (ipy == ipsite(2)) .and. &
 &     (ipz == ipsite(3)) .and. &
 &     (ieo == ye%ieo   ) ) then

    do is=1,SPIN
    do ic=1,COL
      ye%s(itb,iz,iy,ix)%y(ic,is) = Z1
    enddo
    enddo

    write(*,'("# set source at (itx,ity,itz,itt)=",4I3)')itx,ity,itz,itt
 
  endif

  call copy_boundary(ye)

#ifdef _SF
  call set_sf_boundary(ye)
#endif

  return
end subroutine

subroutine get_sf_scheme_coupling_quark(this,gauge,uorg)
!
!  Calculate quark contribution to the SF scheme coupling
!
!  Sq = - N_f Tr[Log[D_W[U]]]
!
! d S_q|                 1
! -----|        = k ( -------- - mu v )
! d eta|_eta=0        (g_SF)^2
!
!     k : knorm : normalization contant depending on the gauge action
!
!  this : SF parameters
! knorm : normaliztion from the gauge action
!  uorg : link field
!
  use comlib,          only : comlib_sumcast
  use hmc_gluon_class, only : gluon_rg_sf
  implicit none
  type(sf_measurement),  intent(inout) :: this
  type(gluon_rg_sf),     intent(in)    :: gauge
  type(vfield_gluon_wg), target, intent(inout) :: uorg

  type(vfield_gluon_wg), allocatable :: BB
  type(tfield_gluon_wg), allocatable :: XX
  type(field_quark_wg),  allocatable :: z2
  type(field_quark_wg),  allocatable :: fz,fy
  type(tfield_gluon_wg), allocatable :: wl34
  type(tfield_gluon_wg), allocatable :: wl98
  type(vfield_gluon_wg), allocatable, target :: uc
  real(DP) :: knorm
  complex(DP) :: zcoef
  real(DP) :: dSQg , dSQv
  real(DP) :: gSF2invq,vSFq
  real(DP) :: tol,dummy,rcoef
  real(DP) :: tmp_gsfq , tmp_vsfq
  integer :: jc,js
  integer :: iout,iter,inoise,itraj
  type(qprop_params), target :: params
  character(len=CHARLEN) :: str
#ifdef _STOUT
  type(stout_smearing_parameters), target :: stout
#endif
  type(vfield_gluon_wg), pointer :: utmp
  integer :: itx,ity,itz,itt

#ifdef _CLOVER

#ifdef _CHECK_TRACE
  allocate(uc)
  call new(uc)
  call set_sf_classical_background(uc)

#ifdef _CHECK_GAUGEINV
  call transform_gauge_sf(uc)
#endif
#endif

  !
  ! gauge action normalizetion knorm
  !   from action on a classical SF field
  !
  knorm = get_gsf_normalization(gauge) 

#ifdef _STOUT
  !-----------------
  ! stout smearing
  !-----------------
  if (0 < this%stout_nstep) then
    if ( 0 == nodeid ) then
      write(*,'("# Link Stout smeared")')
    endif
    call new(stout,this%stout_alpha,this%stout_nstep)
    call smear(stout,uorg)
    utmp => stout%usmr(this%stout_nstep)
  else
    utmp => uorg
  endif
#else
  utmp => uorg
#endif

#ifdef _CHECK_TRACE
  utmp => uc
#endif

  call new(params)
  allocate(quark_clover::params%mass)

  select type(p => params%mass)
  type is (quark_clover)

    call new(p,id=0)
    call set_kappa(p,this%kappa)
    call set_csw(p,this%csw)
    call set_cttilde(p,this%cttilde)

    params%tol        =  this%tolerance
    params%solver_log => this%logfile_gsfq_solver

    allocate(wl34,wl98)
    call new(wl34)
    call new(wl98)

    call print(p)

    !==============================================
    ! compute clover term and two links
    !==============================================
    call make_two_links(utmp,wl34,wl98)
    call make_clover_leaf(utmp,wl34,wl98)
    call make_inverse_clover_term(p,0,dummy)
    call delete_clover_leaf()

    allocate(BB,XX)
    call new(BB)
    call new(XX)

    allocate(z2)
    call new(z2)

    allocate(fz,fy)
    call new(fz)
    call new(fy)

#ifdef _CHECK_TRACE
    do itx = 1,NTX
    do ity = 1,NTY
    do itz = 1,NTZ
    do itt = 2,NTT  ! quark physical domain is defined in itt=2...NTT (t=1..T-1)

#ifdef _CHECK_RAND
      if (0 == nodeid) then
        write(*,'("# TRACE CHECK WITH STOCHASTIC METHOD (Z2-RANDOM FIELD).")')
      endif
      call set_z2_noise(z2)
#else
      if (0 == nodeid) then
        write(*,'("# TRACE CHECK USING COMPLETE COORDINATE BASIS.")')
      endif
      call set_coordinate_basis(z2,itx,ity,itz,itt)
#endif

#else
    do inoise=1,this%random_gsfq%nnoise

      call set_z2_noise(z2)
#endif

      dSQg = 0.0_DP
      dSQv = 0.0_DP

      !
      ! color spin trace
      !
      do js=1,SPIN
      do jc=1,COL

        call clear_all(BB)
        call clear_all(XX)

        write(str,'("# SOLVER FOR COL=",I2," SPIN=",I2)')jc,js
        write(str,'("# Kappa_=",F18.14," csw_=",F18.14," cttilde_=",F18.14)')this%kappa,this%csw,this%cttilde
        call print(this%logfile_gsfq_solver,TRIM(str))
        iout = get_file_unit(this%logfile_gsfq_solver)
        tol  = this%tolerance

        call set_z2_noise_source(fz,z2,jc,js)

!        call dump(fz)

        call assign_inv_mult_wd(p,iout,tol,iter,fy,fz,utmp)

        call copy_boundary(fy)
#ifdef _SF
        call set_sf_boundary(fy)
#endif

#ifdef _CHECK_TRACE
        if ( 0 == nodeid ) then
          write(*,'("@|z2|^2_=",E24.15," site_xyzt_=",4I3," |fz|^2=",E24.15," |fy|^2=",E24.15," jc,js=",2I2)') &
 &        abs2(z2),itx,ity,itz,itt,abs2(fz),abs2(fy),jc,js
        endif
#endif

        rcoef = this%kappa
        call force_hmc_hopping(BB,rcoef,fz,fy)

        zcoef = ZI*this%csw*this%kappa/8.0_DP
        call force_hmc_clover(XX,zcoef,fy,fz)
        call get_gsf_clover_terms(BB,XX,utmp)  ! staple containing XX terms


#ifdef _STOUT
        if (0 < this%stout_nstep) then
          call get_gsf_coupling_contribution(stout,BB,uorg)
        endif
#endif

        call get_sf_scheme_coupling_quark_dot(tmp_gsfq,tmp_vsfq,BB,uorg)

        dSQg = dSQg + tmp_gsfq
        dSQv = dSQv + tmp_vsfq

#ifdef _CHECK_TRACE
        if ( 0 == nodeid ) then
          write(*,'("@tmp_gsfq_=",E24.15," tmp_vsfq_=",E24.15," col_=",I2," spin_=",I2)')tmp_gsfq,tmp_vsfq,jc,js
        endif
#endif

      enddo
      enddo

      gSF2invq =  dSQg/knorm
      vSFq     = -dSQv/knorm
    
      itraj = get_trajectory_number(this%status)
      if (0 == nodeid) then
#ifdef _CHECK_TRACE
        write(str,'(I6,5E24.15,4I3)')itraj,gSF2invq,vSFq,dSQg,-dSQv,knorm,itx,ity,itz,itt
#else
        write(str,'(I6,5E24.15,I9)') itraj,gSF2invq,vSFq,dSQg,-dSQv,knorm,inoise
#endif
        call print(this%logfile_gsfq,TRIM(str))
      endif

#ifdef _CHECK_TRACE
    enddo
    enddo
    enddo
    enddo ! end of do itt,itz,ity,itx
#else
    enddo ! end of do inoise
#endif

    deallocate(fz,fy)
    deallocate(z2)
    deallocate(BB,XX)
    deallocate(wl34,wl98)
    call delete_clover_leaf()

  end select

#ifdef _CHECK_TRACE
  deallocate(uc)
#endif

#ifdef _STOUT
  if (0 < this%stout_nstep) then
    call delete(stout)
  endif
#endif

  call delete(params)
  deallocate(params%mass)

#endif

  return
end subroutine

subroutine get_sf_scheme_coupling_quark_dot(dSQg,dSQv,BB,u)
!
! Single flavor quark contribution to the SF coupling
!
! dSQg  =  dSq/deta |_{eta=0,mu=0}        / Nf
! dSQv  = (dSq/deta |_{eta=0} - dSQg)/mu  / Nf
!
  implicit none
  real(DP), intent(out) :: dSQg, dSQv
  type(vfield_gluon_wg), intent(in) :: BB
  type(vfield_gluon_wg), intent(in) ::  u
  real(DP) :: gsfe,vsfe
  real(DP) :: gsfo,vsfo
  call get_sf_scheme_coupling_quark_dot_eo(gsfe,vsfe,BB%eo(0),u%eo(0))
  call get_sf_scheme_coupling_quark_dot_eo(gsfo,vsfo,BB%eo(1),u%eo(1))
  dSQg = gsfe + gsfo
  dSQv = vsfe + vsfo
  return
end subroutine

subroutine get_sf_scheme_coupling_quark_dot_eo(gsfe,vsfe,BBe,ue)
!
! Single flavor quark contribution to the SF coupling, from even/odd sites
!
!
!  gsfe = sum_ne sum_{k=1,2,3} (1/N_k) ReTr[ i tomg W_k B_k(ne,t=0) - W_k' B_k(ne,t=T) ]
!  vsfe = sum_ne sum_{k=1,2,3} (1/N_k) ReTr[ i teee W_k B_k(ne,t=0) + W_k' B_k(ne,t=T) ]
!
!  tomg = diag(1,-1/2,-1/2),  teee = diag(0,1,-1)
!
!  W_k  = U_k(n,t=0)
!  W_k' = U_k(n,t=T)
!  B_k(n) :  dSq/deta response to dU_k(n)/deta
!
!  gsfe : eta-derivative of quark action on even/odd sites
!  vsfe : eta-derivative of quark action (nu dependent part) on even/odd sites
!   BBe : eta-derivative of quark action, response to dU_mu(n)/deta on even/odd sites
!    ue : gauge link field on even/odd sites
!
  use comlib, only : comlib_sumcast
  implicit none
  real(DP), intent(out) :: gsfe,vsfe
  type(vfield_gluon_eo_wg), intent(in) :: BBe
  type(vfield_gluon_eo_wg), intent(in) ::  ue
  character(len=CHARLEN) :: str
  complex(DP) :: zgsf0,zgsf1
  complex(DP) :: zvsf0,zvsf1
  complex(DP) :: w0(COL,COL)
  complex(DP) :: w1(COL,COL)
  real(DP) :: tomg(COL,NDIM-1),teee(COL,NDIM-1)
  real(DP) :: rtmp
  integer :: ix,iy,iz,itb,ieoxyz
  integer :: ic,jc,mu
  integer :: size_lattice(NDIM)

  if (.not. (ue%ieo == BBe%ieo)) then
    write(str,'("even/odd error stop at get_sf_coupling_quark_dot_eo:",A,I5)')__FILE__,__LINE__
  endif

  size_lattice(1) = NTX
  size_lattice(2) = NTY
  size_lattice(3) = NTZ
  size_lattice(4) = NTT

  do mu=1,NDIM-1
    tomg(1,mu) =  1.0_DP/size_lattice(mu)
    tomg(2,mu) = -0.5_DP/size_lattice(mu)
    tomg(3,mu) = -0.5_DP/size_lattice(mu)
    teee(1,mu) =  0.0_DP/size_lattice(mu)
    teee(2,mu) =  1.0_DP/size_lattice(mu)
    teee(3,mu) = -1.0_DP/size_lattice(mu)
  enddo

  zgsf0 = Z0
  zgsf1 = Z0
  zvsf0 = Z0
  zvsf1 = Z0

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,mu,ic,jc,rtmp,w0,w1) REDUCTION(+:zgsf0,zgsf1,zvsf0,zvsf1) _CLPS_
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ue%ieo,2)
    if (1 == ieoxyz) then
      do mu=1,NDIM-1 ! spatial only

        !========================
        ! itt = 1 (nt=0)
        !========================
        !
        !  w0 = U_k(n,nt=0) * BB_k(n,nt=0)
        !
!        rtmp = 0.0_DP
!        do jc=1,COL
!        do ic=1,COL
!          rtmp = rtmp + REAL(BBe%mu(mu)%s(  0,iz,iy,ix)%u(ic,jc))**2  &
! &                   + AIMAG(BBe%mu(mu)%s(  0,iz,iy,ix)%u(ic,jc))**2 
!        enddo
!        enddo
!        if ( rtmp > EPSILON(1.0_DP)**2*10 ) then
!          do jc=1,COL
!          do ic=1,COL
!            write(*,'("@mu=",I2," ieo=",I2," zyx=",3I3," ue(0),BBe(0)=",4E24.15)') &
! &              mu,ue%ieo,iz,iy,ix,ue%mu(mu)%s(  0,iz,iy,ix)%u(ic,jc),BBe%mu(mu)%s(  0,iz,iy,ix)%u(ic,jc)
!          enddo
!          enddo
!        endif
 
        do jc=1,COL
        do ic=1,COL
          w0(ic,jc) = ue%mu(mu)%s(  0,iz,iy,ix)%u(ic, 1)*BBe%mu(mu)%s(  0,iz,iy,ix)%u( 1,jc) &
 &                  + ue%mu(mu)%s(  0,iz,iy,ix)%u(ic, 2)*BBe%mu(mu)%s(  0,iz,iy,ix)%u( 2,jc) &
 &                  + ue%mu(mu)%s(  0,iz,iy,ix)%u(ic, 3)*BBe%mu(mu)%s(  0,iz,iy,ix)%u( 3,jc)
        enddo
        enddo

        zgsf0 = zgsf0 + ZI*(  tomg(1,mu)*w0(1,1) &
 &                          + tomg(2,mu)*w0(2,2) &
 &                          + tomg(3,mu)*w0(3,3) )
        zvsf0 = zvsf0 + ZI*(  teee(1,mu)*w0(1,1) &
 &                          + teee(2,mu)*w0(2,2) &
 &                          + teee(3,mu)*w0(3,3) )

        !========================
        ! itt = NTT+1 (nt=T)
        !========================
        !
        !  w1 = U_k(n,nt=T) * BB_k(n,nt=T)
        !
!        rtmp = 0.0_DP
!        do jc=1,COL
!        do ic=1,COL
!          rtmp = rtmp + REAL(BBe%mu(mu)%s(NTH,iz,iy,ix)%u(ic,jc))**2  &
! &                   + AIMAG(BBe%mu(mu)%s(NTH,iz,iy,ix)%u(ic,jc))**2 
!        enddo
!        enddo
!        if ( rtmp > EPSILON(1.0_DP)**2*10 ) then
!          do jc=1,COL
!          do ic=1,COL
!            write(*,'("@mu=",I2," ieo=",I2," zyx=",3I3," ue(T),BBe(T)=",4E24.15)') &
! &              mu,ue%ieo,iz,iy,ix,ue%mu(mu)%s(NTH,iz,iy,ix)%u(ic,jc),BBe%mu(mu)%s(NTH,iz,iy,ix)%u(ic,jc)
!          enddo
!          enddo
!        endif

        do jc=1,COL
        do ic=1,COL
          w1(ic,jc) = ue%mu(mu)%s(NTH,iz,iy,ix)%u(ic, 1)*BBe%mu(mu)%s(NTH,iz,iy,ix)%u( 1,jc) &
 &                  + ue%mu(mu)%s(NTH,iz,iy,ix)%u(ic, 2)*BBe%mu(mu)%s(NTH,iz,iy,ix)%u( 2,jc) &
 &                  + ue%mu(mu)%s(NTH,iz,iy,ix)%u(ic, 3)*BBe%mu(mu)%s(NTH,iz,iy,ix)%u( 3,jc)
        enddo
        enddo

        zgsf1 = zgsf1 - ZI*(  tomg(1,mu)*w1(1,1) &
 &                          + tomg(2,mu)*w1(2,2) &
 &                          + tomg(3,mu)*w1(3,3) )
        zvsf1 = zvsf1 + ZI*(  teee(1,mu)*w1(1,1) &
 &                          + teee(2,mu)*w1(2,2) &
 &                          + teee(3,mu)*w1(3,3) )

      enddo
    endif
  enddo
  enddo
  enddo  ! end of do ix,iy,iz

#ifndef _singlePU
  call comlib_sumcast(zgsf0)
  call comlib_sumcast(zvsf0)
  call comlib_sumcast(zgsf1)
  call comlib_sumcast(zvsf1)
#endif

!  if ( 0 == nodeid ) then
!    write(*,'("@zgsf0_=",2E24.15," zgsf1_=",2E24.15)')zgsf0,zgsf1
!    write(*,'("@zvsf0_=",2E24.15," zvsf1_=",2E24.15)')zvsf0,zvsf1
!  endif

  gsfe = REAL(zgsf0,kind=DP) + REAL(zgsf1,kind=DP)
  vsfe = REAL(zvsf0,kind=DP) + REAL(zvsf1,kind=DP)

  return
end subroutine

#include "get_gsf_clover_terms.h90"
#include "set_sf_classical_bg.h90"

end module
