#include "../config.h"
  
!#define _TEST_DUMPY_
#undef _TEST_DUMPY_

module reweighting_ovlp_class
  use comlib
  use constants_module
  use random_class
  use error_class
  use counter_class
  use timer_class
  use lattice_class
  use logfile_class
  use print_status_class
  use field_gauge_class
  use field_fermion_class

  use quark_dw_ovlp_class
  use reweighting_truncd_overlap_class

  use config_profile_class

  use quark_prop_class, only : assign_inv_mult_wd
  implicit none
  private

  public :: reweighting_ovlp
  public :: new,delete
  public :: read,save,print,print_statistics,run

  integer, parameter :: MAX_IR_SOLVER_ITER=10

  type reweighting_ovlp
    type(vfield_gluon_wg), allocatable :: u
    type(config_profile_obj)   :: config_prof

    type(reweighting_truncd_dw_overlap) :: reweight

    type(timer)   :: tot_time, io_time, run_time

    real(DP)  :: stout_alpha = 0.0_DP
    integer   :: stout_nstep = 0

    integer :: run=0
    integer :: traj_start=0
    integer :: traj_end=0
    integer :: traj_skip=0
    integer :: traj=0
    integer :: random_number_seed=1

    character(len=CHARLEN) :: input_fname=""
    character(len=CHARLEN) :: cwd=""
    character(len=CHARLEN) :: job_id=""
    character(len=CHARLEN) :: prog_version=""
    character(len=CHARLEN) :: ctime=""
    character(len=CHARLEN) :: cdate=""
    character(len=CHARLEN) :: commentline(6)=["","","","","",""]
    character(len=CHARLEN) :: reweight_fname=""
    character(len=CHARLEN) :: reweight_file="reweight_hist" ! reweight filename
    character(len=CHARLEN) :: solver_file="solver_log"      ! reweight filename
    character(len=CHARLEN) :: config_path=""                ! configuration file path
    character(len=CHARLEN) :: config_file=""                ! configuration file name
    character(len=CHARLEN) :: seed_file="ReweightRandSeed"  ! seed filename
    character(len=CHARLEN) :: seed_path=""                  ! seed path

  end type

  type reweighting_factor
    real(DP) :: rho,dS,dSA,dSB,dSC
  end type

  interface new
    module procedure new_reweighting
  end interface

  interface delete
    module procedure delete_reweighting
  end interface

  interface read
    module procedure read_reweighting
  end interface

  interface print
    module procedure print_reweighting
  end interface

  interface print_statistics
    module procedure print_stat_reweighting
  end interface

  interface save
    module procedure save_reweighting
  end interface

  interface run
    module procedure run_reweighting
  end interface

contains

subroutine new_reweighting(this)
  use hmc_identity
  implicit none
  type(reweighting_ovlp), intent(inout) :: this
  type(lattice_world)   :: lattice
  integer :: id

  call new(lattice)

  call new(this%tot_time)
  call tic(this%tot_time)

  call new(this%io_time)
  call new(this%run_time)

  !=====================================
  ! Initialize Status
  !=====================================
  this%input_fname  = repeat(' ',len(this%input_fname))
  this%cwd          = repeat(' ',len(this%cwd))
  this%job_id       = repeat(' ',len(this%job_id))
  this%prog_version = repeat(' ',len(this%prog_version))

  call DATE_AND_TIME(date=this%cdate,time=this%ctime)
  write(this%job_id,'(A,1X,A)')TRIM(ADJUSTL(this%cdate)),TRIM(ADJUSTL(this%ctime))
  
  this%prog_version = TRIM(hmc_vers)

  if (nodeid == 0) then
    call GET_COMMAND_ARGUMENT(1,this%input_fname)
    call getcwd(this%cwd)
  endif

  if (nodeid == 0) then
    write(*,'(" DATE AND TIME :",A)')TRIM(this%job_id)
    write(*,'(A)')TRIM(this%prog_version)
  endif

  if (NPU > 1) then
    call comlib_bcast(this%input_fname,0)
    call comlib_bcast(this%cwd,0)
    call comlib_bcast(this%job_id,0)
    call comlib_bcast(this%ctime,0)
    call comlib_bcast(this%cdate,0)
  endif

  allocate(this%u)
  call new(this%u)


  call new(this%config_prof)


  return
end subroutine

subroutine delete_reweighting(this)
  implicit none
  type(reweighting_ovlp), intent(inout) :: this
  type(lattice_world)                   :: lattice

!  call delete(this%reweight)

  call delete(this%config_prof)

  deallocate(this%u)

  call delete(lattice)

  return
end subroutine

subroutine read_reweighting(this)
  use file_tools_class
  use hmc_status_class, only : fname_cont
  implicit none
  type(reweighting_ovlp), intent(inout) :: this
  integer :: iout
  character(len=CHARLEN) :: char,fname,str
  real(DP) :: target_kappa,target_csw
  real(DP) :: config_kappa,config_csw

  call tic(this%io_time)

  iout = search_free_file_unit()
!=====================================
! Read Input Parameters
!=====================================
  if (nodeid==0) then

    open(iout,file=this%input_fname,status='unknown',form='formatted')

    read(iout,'(A)')this%commentline(1) ! skip comment line

    read(iout,'(A)')this%reweight_fname

    read(iout,'(A)')this%commentline(2) ! skip comment line

    read(iout,'(A)')this%config_path
    read(iout,'(A)')this%config_file
    read(iout,*)this%stout_alpha,this%stout_nstep

    read(iout,'(A)')this%commentline(3) ! skip comment line

    read(iout,'(A)')this%seed_path
    read(iout,'(A)')this%seed_file
    read(iout,*)this%random_number_seed

    read(iout,'(A)')this%commentline(4) ! skip comment line

    read(iout,*)this%run
    read(iout,*)this%traj_start,this%traj_end,this%traj_skip

    close(iout)
  endif


#ifndef _STOUT
  this%stout_alpha = 0.0_DP
  this%stout_nstep = 0
#endif

#ifndef _singlePU
  call comlib_bcast(this%stout_alpha,0)
  call comlib_bcast(this%stout_nstep,0)
  call comlib_bcast(this%config_path,0)
  call comlib_bcast(this%config_file,0)
  call comlib_bcast(this%seed_path,0)
  call comlib_bcast(this%seed_file,0)
  call comlib_bcast(this%run,0)
  call comlib_bcast(this%traj_start,0)
  call comlib_bcast(this%traj_end,0)
  call comlib_bcast(this%traj_skip,0)
  call comlib_bcast(this%random_number_seed,0)
#endif


!=====================================
! Check Input Parameters
!=====================================
  if (this%traj_end < this%traj_start) then
    write(str,'(" ERROR: traj_start > traj_end !")')
    call error_stop(str)
  endif
  if (this%run < 0) then
    write(str,'(" ERROR: run < 0 !")')
    call error_stop(str)
  endif
  if (nodeid == 0) then
    write(*,'(80("-"))')
  endif
  write(char,'(1X,A," : read OK ",I3)')TRIM(this%input_fname),nodeid
  call print_status(char)

  call new(this%reweight,id=this%run+1,is_hmc=.FALSE.)

  !------------------------------------------
  ! read reweighting job parmeters
  !------------------------------------------
  if (nodeid==0) then
    iout = search_free_file_unit()
    open(iout,file=TRIM(this%reweight_fname),status='unknown',form='formatted')
  endif

  call read(this%reweight,iout)

  if (nodeid==0) then
    close(iout)
  endif

!=====================================
! Initialize Random Number generator
!=====================================
  fname=fname_cont(this%run,this%seed_path,this%seed_file)
  select case (this%run)
  case (0)
    call new(g_rand,0,this%random_number_seed,NPU,nodeid)
    if (nodeid == 0) then
      write(*,'(80("-"))')
      write(*,'(1X,"Initialize Random Number")')
    endif
  case (1:)
    call new(g_rand,1,this%random_number_seed,NPU,nodeid)
    open(iout,file=fname,status='unknown',form='unformatted',action='read')
    call read(g_rand,iout)
    close(iout)
    if (nodeid == 0) then
      write(*,'(80("-"))')
    endif
    write(char,'(1X,A," : read OK ",I3)')TRIM(fname),nodeid
    call print_status(char)
  end select

  call toc(this%io_time)

  return
end subroutine

subroutine print_reweighting(this)
  implicit none
  type(reweighting_ovlp), intent(inout) :: this
  character(len=CHARLEN) :: cahr

  call tic(this%io_time)
  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'("***** Reweighting Paramtere file :",(A))')TRIM(this%reweight_fname)
    call print(this%reweight)
    write(*,'(80("-"))')
    write(*,'("            Config Path :",A)')TRIM(this%config_path)
    write(*,'("            Config File :",A)')TRIM(this%config_file)
    write(*,'("              Seed Path :",A)')TRIM(this%seed_path)
    write(*,'("              Seed File :",A)')TRIM(this%seed_file)
    write(*,'(80("-"))')
    write(*,'("         Job Parameters :")')
    write(*,'("              Rand Seed :",I12)')this%random_number_seed
    write(*,'("                  Run # :",I4)')this%run
    write(*,'(" Traj. Start, End, Skip :",2I6,I4)')this%traj_start,this%traj_end,this%traj_skip
    write(*,'(80("="))')
  endif
  call toc(this%io_time)

  return
end subroutine

subroutine run_reweighting(this)
#ifdef _STOUT
  use hmc_stout_class
#endif
  implicit none
  type(reweighting_ovlp), intent(inout) :: this
#ifdef _STOUT
  type(stout_smearing_parameters) :: stout
#endif
  integer :: itraj,inoise

#ifdef _CLOVER
  type(tfield_gluon_wg), allocatable :: wl34,wl98
  real(DP) :: dummy
#endif

  type(reweighting_factor) :: rfactor
  real(DP) :: plq0
  real(DP) :: dS,rho,rtmp,dSA,dSB,dSC
  complex(DP) :: ctmp
  character(len=CHARLEN) :: str

  character(len=CHARLEN) :: fbuff
  character(len=CHARLEN) :: config_path2

  this%run=this%run+1



  do itraj=this%traj_start,this%traj_end,this%traj_skip

!=====================================
! Read Measurment Configuration
!=====================================
    call read_config_measurement(itraj,this%config_path,this%config_file,this%config_prof,this%u)

#ifdef _STOUT
!-------------------
!  Stout smearing
!-------------------
    if (this%stout_alpha > 0.0_DP .and. this%stout_nstep > 0 ) then
      call new(stout,alpha=this%stout_alpha,nstout=this%stout_nstep,is_smear_only=.TRUE.)
      call print(stout)
      call smear(stout,this%u)
      call assign(this%u,stout%usmr(1))
      call delete(stout)
      call copy_boundary(this%u)
      call plaquette(this%u,plq0)
      if (nodeid==0) write(*,'(" PLQ(STOUT):",F24.16)')plq0
    endif
#endif

    call run(this%reweight,itraj,this%u)

  enddo

  call delete(this%reweight)

  return
end subroutine


subroutine save_reweighting(this)
  use file_tools_class
  use hmc_status_class, only : fname_cont
  implicit none
  type(reweighting_ovlp), intent(inout) :: this
  integer :: iout
  character(len=CHARLEN) :: fname
  real(DP) :: config_kappa,config_csw
  real(DP) :: target_kappa,target_csw

  call tic(this%io_time)

  !=====================================
  ! Save Random Seed
  !=====================================
  fname=fname_cont(this%run,this%seed_path,this%seed_file)
  iout = search_free_file_unit()
  open(iout,file=fname,status='unknown',form='unformatted',action='write')
  call save(g_rand,iout)
  close(iout)


  !=====================================
  ! Save Input Parameters
  !=====================================
  if (nodeid==0) then
    open(iout,file=this%input_fname,status='unknown',form='formatted')

    write(iout,'(A)')TRIM(this%commentline(1)) ! skip comment line

    write(iout,'(A)')TRIM(this%reweight_fname)

    write(iout,'(A)')TRIM(this%commentline(2)) ! skip comment line

    write(iout,'(A)')TRIM(this%config_path)
    write(iout,'(A)')TRIM(this%config_file)
    write(iout,'(E14.6,I3)')this%stout_alpha,this%stout_nstep

    write(iout,'(A)')TRIM(this%commentline(3)) ! skip comment line

    write(iout,'(A)')TRIM(this%seed_path)
    write(iout,'(A)')TRIM(this%seed_file)
    write(iout,*)this%random_number_seed

    write(iout,'(A)')TRIM(this%commentline(4)) ! skip comment line

    write(iout,*)this%run
    write(iout,*)this%traj_start,this%traj_end,this%traj_skip

    close(iout)
  endif
  call toc(this%io_time)

  return
end subroutine

subroutine print_stat_reweighting(this)
  implicit none
  type(reweighting_ovlp), intent(inout) :: this
  real(DP) :: run_time,io_time,tot_time,iter_ave

!  iter_ave = get_average(this%solver_iter)

  run_time = get_elapse(this%run_time)
   io_time = get_elapse(this%io_time)
  call toc(this%tot_time)
  tot_time = get_elapse(this%tot_time)

  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'(" Rweighting Truncated-DW-Overlap")')
  endif

  call print_statistics(this%reweight)

  if (nodeid==0) then
    write(*,'(80("-"))')

    write(*,'(" Total Time :",F24.15)')tot_time
    write(*,'("    IO Time :",F24.15," (",F12.6," %)")')  io_time,100* io_time/tot_time
    write(*,'("   RUN Time :",F24.15," (",F12.6," %)")') run_time,100*run_time/tot_time
!    write(*,'("Solver Iter :",F14.7)')iter_ave
    write(*,'(80("="))')
  endif

  return
end subroutine


#include "read_config.h90"

end module
